import pandas as pd
import numpy as np
#import seaborn as sns
import matplotlib.pyplot as plt
import Bio.Seq

def mutation_aggregator(mutant_input, gene_name):    
    """
    Selects the mutations that correspond to the desired gene name.
    Removes duplicates (checking at DNA level; not amino acid level).
    Returns dataframe containing these aggregated mutants occuring in gene_name.
    
    Parameters
    ----------
    mutant_input
        *type = pd.DataFrame*
        
        A dataframe containing the input mutations from which selections are made to generate pegRNAs.
        See documentation for precise qualities of this dataframe
        
    gene_name
        *type = str*
        
        Gene's Hugo_Symbol (i.e. name).
        
    """
    
    gene_mutants = mutant_input[mutant_input['Hugo_Symbol']==gene_name]
    mutant_input_sparse = gene_mutants[["Hugo_Symbol", "Chromosome", "Start_Position", "End_Position", "Variant_Type", "Reference_Allele", "Tumor_Seq_Allele2"]].drop_duplicates()
    mutant_idxs = list(mutant_input_sparse.index)
    
    return mutant_input.iloc[mutant_idxs]


def neutral_substitutions(gene_name, chrom, strand, start_end_cds, chrom_dict):
    """
    A function for generating all possible synonymous codon substitutions (i.e. silent mutations) of a given gene.
    See documentation for more information about format of start_end_cds & example usage.

    Parameters
    ----------
    gene_name
        *type = str*
        
        Gene's Hugo_Symbol (i.e. name).
        
    chrom
        *type = str*
        
        Chromosome that gene occurs on. Format = e.g. 'chr17', 'chrX', etc.
        
    strand
        *type = str*
        
        Strand that transcript is on. Options are '+' or '-'.
        
    start_end_cds
        *type = list*
        
        A 2-d list containing the start/end locations of each region of the coding sequence (CDS) for the gene's selected transcript.
        See documentation for example and precise specifications of format. 
        
    chrom_dict
        *type = dict*
        
        Dictionary containing the reference genome. See genome_loader() 
    
    """
    
    
    codons = []
    for i in start_end_cds:
        for k in range(i[0], i[1]+1):
            codons.append(k)

    gene_codons = []
    if strand=='+':
        for i in range(0,len(codons),3):
            gene_codons.append([codons[i], codons[i+1], codons[i+2]])

    elif strand=='-':
        codons = codons[::-1]
        for i in range(0,len(codons),3):
            gene_codons.append([codons[i], codons[i+1], codons[i+2]])
    
    
    #-------------------creating df of codons  
    
    #loading in reference genome from peg engine module
    seq1 = chrom_dict[chrom]

    codons = []
    for i in gene_codons:
        s1 = seq1[i[0]-1]+seq1[i[1]-1]+seq1[i[2]-1]
        se = Bio.Seq.Seq(s1)
        codons.append(str(se))


    codon_start = []
    codon_end = []
    ref_codon = []
    hugo_symbol = [gene_name]*len(gene_codons)
    chrom1 = [chrom]*len(gene_codons)
    variant_type = ['ONP']*len(gene_codons)
    strand1 = [strand]*len(codons)

    for idx, val in enumerate(gene_codons):

        if strand=='-':
            ref_codon.append(codons[idx][::-1]) #need it in + strand orientation

        else:
            ref_codon.append(codons[idx]) #need it in + strand orientation


        s = val[2] #start = end
        e = val[0]
        codon_start.append(s)
        codon_end.append(e)


    g_cod = pd.DataFrame(data = hugo_symbol, columns = ['Hugo_Symbol'])
    g_cod['Chromosome']=chrom1
    g_cod['Start_Position'] = codon_start
    g_cod['End_Position']=codon_end
    g_cod['Variant_Type'] = variant_type
    g_cod['Reference_Allele'] = ref_codon
    g_cod['Tumor_Seq_Allele2']=ref_codon #change to mutations later
    g_cod['codon']=range(1,len(gene_codons)+1)

    #it's a minus strand protein
    if strand=='-':

        codon_list = list(g_cod['Reference_Allele'])
        codon_list_rc = [str(Bio.Seq.translate(Bio.Seq.transcribe(Bio.Seq.Seq(i).reverse_complement()))) for i in codon_list]

    else:

        codon_list = list(g_cod['Reference_Allele'])
        codon_list_rc = [str(Bio.Seq.translate(Bio.Seq.transcribe(Bio.Seq.Seq(i)))) for i in codon_list]


    g_cod['ref_aa']=codon_list_rc
    
    
    #------------------enumerate all synonymous codon substitutions at each position
    all_codons = []
    bases = ['A','T','C','G']
    for i in bases:
        for k in bases:
            for j in bases:
                new = i+k+j
                all_codons.append(new)


    #translations for all codons
    aas=[str(Bio.Seq.Seq(i).transcribe().translate()) for i in all_codons]
    unique_aas = list(np.unique(aas))

    #creating list corresponding to codons for each amino acid
    matched_codons = [[] for x in range(len(unique_aas))]
    index_list = []
    for idx, val in enumerate(all_codons):
        aa = str(Bio.Seq.Seq(val).transcribe().translate()) #determine amino acid from codon
        index=unique_aas.index(aa) #find index of aa in unique_aa list
        matched_codons[index].append(val)


    #------------------now creating one synonymous subtitution for each aa (iff possible)
    tumor_allele = []
    mut_aa = []
    for idx, val in enumerate(list(g_cod['ref_aa'])):

        ref_cod = g_cod.iloc[[idx]]['Reference_Allele'].values[0]


        #finding corresponding codons
        cod_idx = unique_aas.index(val)
        possible_codons = matched_codons[cod_idx]

        if strand=='-':

            possible_codons = [str(Bio.Seq.Seq(i).reverse_complement()) for i in possible_codons]

            if len(possible_codons)==1:
                tumor_allele.append(possible_codons[0])
                mut_aa.append(str(Bio.Seq.Seq(possible_codons[0]).reverse_complement().transcribe().translate()))

            else:
                if possible_codons[0]==ref_cod:
                    tumor_allele.append(possible_codons[1])
                    mut_aa.append(str(Bio.Seq.Seq(possible_codons[1]).reverse_complement().transcribe().translate()))
                    
                else:
                    tumor_allele.append(possible_codons[0])
                    mut_aa.append(str(Bio.Seq.Seq(possible_codons[0]).reverse_complement().transcribe().translate()))


        else:

            if len(possible_codons)==1:
                tumor_allele.append(possible_codons[0])
                mut_aa.append(str(Bio.Seq.Seq(possible_codons[0]).transcribe().translate()))


            else:

                if possible_codons[0]==ref_cod:
                    tumor_allele.append(possible_codons[1])
                    mut_aa.append(str(Bio.Seq.Seq(possible_codons[1]).transcribe().translate()))


                else:
                    tumor_allele.append(possible_codons[0])
                    mut_aa.append(str(Bio.Seq.Seq(possible_codons[0]).transcribe().translate()))


    g_cod['Tumor_Seq_Allele2']=tumor_allele
    g_cod['mut_aa']=mut_aa

    #------------------classifying and filtering

    #now classifying mutations according to original and mutant
    classifier = []

    for i, val in g_cod.iterrows():

        o = val['ref_aa']
        m = val['mut_aa']
        o_dna = val['Reference_Allele']
        m_dna = val['Tumor_Seq_Allele2']

        if o_dna == m_dna:
            classifier.append('unchanged')

        else:
            if o==m:
                classifier.append('neutral')
            elif o != m:
                if m=='*':
                    classifier.append('stop')
                else:
                    classifier.append('missense')

    g_cod['classification']=classifier

    g_cod = g_cod[g_cod['classification']=='neutral'].reset_index().drop(columns='index')

    #------and finally fixing the issue where SNPs are labelled as ONPS...

    for i, val in g_cod.iterrows():
        s = val['Start_Position']
        e = val['End_Position']

        r = val['Reference_Allele']
        a = val['Tumor_Seq_Allele2']

        diff_loc = []
        for index, val2 in enumerate(r):
            if val2!=a[index]:
                diff_loc.append(index)

        if len(diff_loc)==1:
            g_cod.loc[i, 'Variant_Type'] = 'SNP'
            g_cod.loc[i, 'Start_Position'] = s + diff_loc[0]
            g_cod.loc[i, 'End_Position'] = s + diff_loc[0]
            g_cod.loc[i, 'Reference_Allele'] = r[diff_loc[0]]
            g_cod.loc[i, 'Tumor_Seq_Allele2'] = a[diff_loc[0]]

    return g_cod


def safe_muts(num_muts, chrom_dict, organism='human'):
    """ 
    Function for generating a list of safe-targetting pegRNAs/gRNAs.
    NOTE: THIS ONLY WORKS FOR GRCh37 and GRCm38!!!
    Alternatively, generate these mutations/guides using these reference builds, and then generate your own guides
    targeting mutations in your desired reference build.

    This is a random subset of 100,000 "safe regions" taken from Morgens et al., 2017 (https://doi.org/10.1038/ncomms15178)

    Parameters
    -----------
    num_muts
        *type = int*
        
        Number of safe-targetting mutations to select/generate.

    chrom_dict
        *type = dict*

        Dictionary containing chromosomes in dictionary format for parsing by PEGG.

    organism
        *type = str*

        Choices = "human" or "mouse". Generates safe targeting guides in human or mouse genome.
    """

    assert organism in ['human', 'mouse'], "Pick a valid reference organism ('human' or 'mouse')"

    if organism == 'human':
        d = pd.read_csv('human_safe_regions_bassik.csv')

    if organism == 'mouse':
        d = pd.read_csv('mouse_safe_regions_bassik.csv')

    #dealing with mix of strings and ints in column because of pandas saving/loading issue with dtypes
    l = list(range(1,23))
    l = [str(i) for i in l]

    #choose regions without replacement
    a = list(range(len(d)))
    rand_choices = np.random.choice(a, size=num_muts, replace=False)
    d = d.iloc[rand_choices].reset_index()

    #now generate fake mutations (same REF and ALT) in cBioPortal format
    loc = []
    chrom_list = []
    dna_seq = []
    for i, val in d.iterrows():
        chrom = val['Chromosome']

        if chrom in l: #fixing int/str issue
            chrom = int(chrom)

        s = val['Start']
        e = val['End']
        #choose a random spot in the safe region as well...
        choice = np.random.randint(s,e)
        dna = chrom_dict[chrom][choice-1].upper() #-1 for 0 based indexing

        loc.append(choice)
        dna_seq.append(dna)
        chrom_list.append(chrom)


    col_labels = [ 'Chromosome', 'Variant_Type', 'Start_Position', 'End_Position', 'Reference_Allele', 'Tumor_Seq_Allele2', 'classification']
    cols = [chrom_list, 'SNP', loc, loc, dna_seq, dna_seq, 'safe-targeting control']
    safe = pd.DataFrame(dict(zip(col_labels, cols)))

    safe = safe.sort_values(by=['Chromosome', 'Start_Position']).reset_index(drop=True)

    return safe


def aavs1_muts(chrom_dict, num_muts, genome_version = 'GRCh37'):
    """ 
    Function for generating list of "mutations" within AAVS1 locus as negative controls.
    Generates a list of SNPs that contain idential reference and alternate alleles.
    Intron 1-2 of PPP1R12C (AAVS1 locus): GRCh37 transcript = ENST00000263433.3 | GRCh38 transcript = ENST00000263433.8
    
    Not currently used in the pipeline, but provided anyway.

    Parameters
    -----------
    chrom_dict
        *type = dict*

        Dictionary containing chromosomes in dictionary format for parsing by PEGG.

    num_muts
        *type = int*
        
        Number of safe-targetting mutations to select/generate.

    genome_version
        *type = str*

        Options = 'GRCh37', 'GRCh38'

    """

    assert genome_version in ['GRCh37', 'GRCh38'], "Pick a valid reference genome ('GRCh37' or 'GRCh38')"

    if genome_version == 'GRCh37':
        #give 500 bp of buffer on either side of the intron
        start_intron = 55624164 + 500
        end_intron = 55628590 - 500

    elif genome_version == 'GRCh38':
        #give 500 bp of buffer on either side of the intron
        start_intron = 55112796 + 500
        end_intron = 55117222 - 500

    chr19 = chrom_dict[19]

    #select a random set of sequence locations in AAVS1
    snp_locs = np.random.random_integers(low = start_intron, high=end_intron, size=num_muts)

    #go through and generate the fake SNPs (same WT and REF allele) for these locations
    dna_seq = []
    for i in snp_locs:
        dna = chr19[int(i)-1]
        dna_seq.append(dna.upper())

    col_labels = ['Hugo_Symbol', 'Chromosome', 'Variant_Type', 'Start_Position', 'End_Position', 'Reference_Allele', 'Tumor_Seq_Allele2', 'classification']
    cols = ['AAVS1', 19, 'SNP', snp_locs, snp_locs, dna_seq, dna_seq, 'AAVS1-targeting control']
    aavs1 = pd.DataFrame(dict(zip(col_labels, cols)))

    return aavs1


def nontargeting_guides(num_guides, edit_type='prime'):
    """ 
    Function for generating a list of non-targetting guides (in the human genome).
    Future versions will expand to mouse as well...Unprocessed files for mouse located within this module.
    Guides taken from: https://www.nature.com/articles/nmeth.4423 (https://doi.org/10.1038/nmeth.4423)

    Parameters
    -----------
    num_guides
        *type = int*

        Number of guides to non-targetting generate. Max = 1000.

    edit_type
        *type = str*

        Options = 'base', 'prime'
    """

    assert edit_type in ['base', 'prime'], "Choose valid edit_type (choices = 'prime', 'base')"

    if num_guides>1000:
        print('Only 1000 non-targeting guides available; will add 1000 guides.')
        num_guides = 1000

    if edit_type == 'prime':
        df = pd.read_csv('non_targeting_human_prime.csv')

    if edit_type == 'base':
        df = pd.read_csv('non_targeting_human_base.csv')

    df = df[:num_guides]

    return df
    


def library_maker(mutant_input, gene_name, chrom_dict, fraction_safetarget=0.05,organism='human', 
                  fraction_silent=0, chrom=None, strand=None, start_end_cds=None):
    """ 
    Compiles the different library design functions to aggregate all of the variants for a particular gene,
    include silent substitution controls at a designated %, and non-targeting controls. This can then be fed
    into the pegg2.run() or base_editing.run_base() functions.

    NOTE: THIS ONLY WORKS FOR GRCh37 and GRCm38!!!
    Alternatively, generate these mutations/guides using these reference builds, and then generate your own guides
    targeting mutations in your desired reference build.

    Parameters
    -----------
    mutant_input
        *type = pd.DataFrame*

        DataFrame containing all of the input mutations. Doesn't require filtration for gene of interest.
    
    gene_name
        *type = str*

        Gene name to select from the mutant_input dataframe. 

    chrom_dict
        *type = dict*

        Dictionary containing the reference genome. See genome_loader() 

    fraction_safetarget
        *type = float*

        Value from 0 to 1 that corresponds to the fraction of safe-targetting mutations to include in the library.
        (i.e. what fraction of the mutant_input)

    organism
        *type = str*

        Options = 'human' or 'mouse'. Determines which of the organisms to generate safe targeting guides for.

    fraction_silent
        *type = float*

        Value from 0 to 1 that corresponds to the fraction of silent mutations to include in the library.
        (i.e. what fraction of the mutant_input)

    chrom
        *type = int or str*

        Chromosome that the gene of interest falls on.

    strand
        *type = str*

        '+' or '-' -- corresponds with which strand the transcript falls on.

    start_end_cds
        *type = list*

        Nested list that contains the CDS locations of the gene transcript of interest, in the + strand orientation.
    
    """
    assert (fraction_silent>=0) & (fraction_silent<=1), "Choose valid fraction_silent (between 0 and 1)"
    assert (fraction_safetarget>=0) & (fraction_safetarget<=1), "Choose valid fraction_nontarget (between 0 and 1)"
    assert organism in ['human', 'mouse'], "Pick a valid reference organism ('human' or 'mouse')"

    #aggregate variants corresponding to gene
    agg_muts = mutation_aggregator(mutant_input, gene_name)
    agg_muts['classification'] = 'variant'
    total_size = len(agg_muts)

    #generate neutral/silent substitution controls
    if fraction_silent>0:
        neutrals = neutral_substitutions(gene_name, chrom, strand, start_end_cds, chrom_dict)

        num_silent = min(int(total_size*fraction_silent), len(neutrals))

        a = list(range(len(neutrals)))
        rand_choices = np.random.choice(a, size=num_silent, replace=False)
        neutrals = neutrals.iloc[rand_choices].reset_index()

        #generate safe-targetting variants
        num_safe = int(total_size*fraction_safetarget)
        safe = safe_muts(num_safe, chrom_dict, organism)

        return pd.concat([agg_muts, neutrals, safe]).reset_index(drop=True)

    else:
        #generate safe-targetting variants
        num_safe = int(total_size*fraction_safetarget)
        safe = safe_muts(num_safe, chrom_dict, organism)

        return pd.concat([agg_muts, safe]).reset_index(drop=True)


#---------Visualization-----(Legacy function from PEGG1.0)

def matrix_rep_library(rf,gene_name,start_end_cds, strand, plot=True):
    #requires adding HGVSp information to SNPs in dataset...
    """
    LEGACY FUNCTION FROM PEGG1.0 (NOT UPDATED)! -- untested; will likely throw errors

    Visualization function for looking at the number of pegRNAs generated in each codon of target gene.
    Also visualizes the identitity of mutant amino acid for SNPs in the input dataframe.
    
    FOR PROPER FUNCTION, REQUIRES USER INCLUSION OF HGVSp INFORMATION IN RF. See documentation for more info.
    
    Returns matrix_snp, matrix_ins, matrix_del, matrix_neutral, fig. First four return objects correspond with
    matrices used for plotting. Fig = figure, which can be saved/exported.
    
    Parameters
    ----------
    rf
        *type = pd.DataFrame*
        
        A dataframe containing the pegRNAs generated by run().
        
    gene_name
        *type = str*
        
        Gene's Hugo_Symbol (i.e. name).
        
    start_end_cds
        *type = list*
        
        A 2-d list containing the start/end locations of each region of the coding sequence (CDS) for the gene's selected transcript.
        See documentation for example and precise specifications of format. 
        
    strand
        *type = str*
        
        Strand that transcript is on. Options are '+' or '-'.
        
    plot
        *type = boolean ('True' or 'False')*
        
        If true, show stacked bar plot. If false, simply return the matrices.
        

    """
    
    
    zero_array = lollipop_library(rf, gene_name, start_end_cds, strand, plot=False)
    
    df = rf
    
    amino_acids = ['ala','arg','asn','asp','cys','glu','gln','gly',
    'his','ile','leu','lys','met','phe','pro','ser',
    'thr','trp','tyr','val','TER']

    amino_acids = [i.upper() for i in amino_acids]
    dictionary = dict(zip(amino_acids, range(0,22)))
    
    
    
    #gene_idx = unique_genes.index(gene_name)
    #codons = codon_locs[gene_idx]
    

    codons = []
    for i in start_end_cds:
        for k in range(i[0], i[1]+1):
            codons.append(k)

    gene_codons = []
    if strand=='+':
        for i in range(0,len(codons),3):
            gene_codons.append([codons[i], codons[i+1], codons[i+2]])

    elif strand=='-':
        codons = codons[::-1]
        for i in range(0,len(codons),3):
            gene_codons.append([codons[i], codons[i+1], codons[i+2]])
    
    gene_codons = np.asarray(gene_codons)
    
    len_gene = len(gene_codons)+1
    
    matrix_snp = np.zeros((len(amino_acids), len_gene))
    matrix_neutral = np.zeros((1, len_gene))
    matrix_ins = np.zeros((1, len_gene))
    matrix_del = np.zeros((1, len_gene))
    
    
    
    
    for i, val in df.iterrows():
        
        mut_class = val['classification']
        
        typ = val['variant_type']
        
        #find codon
        i,j = np.where(gene_codons == val['start'])
        if len(i)>0:
            cod = i[0]
            
        else:
            cod='none'
        
        
        if typ=='SNP':
            
            HGVSp = val['HGVSp']
            if type(HGVSp)==str:
                #extracting relevant information from HGVSp
                if any(i.isdigit() for i in HGVSp)==True:
                    a = HGVSp.split('p.')[1]
                    split_a = re.split(r'(\d+)', a)
                    ref = split_a[0]
                    mut = split_a[2][0:3]
                    cod = int(split_a[1])

                 #if there's a valid HGVSp

                    mut_idx = dictionary[mut.upper()]
                    matrix_snp[mut_idx][int(cod)-1]+= 1#cod-1 corrects for shift by 1
                else:
                    continue
            else:
                continue #no HGVSp provided, continue
          
        
        if type(cod)!=str:
            if typ=='INS':
                matrix_ins[0][cod-1]+=1
            if typ=='DEL':
                matrix_del[0][cod-1]+=1

            if mut_class=='neutral':
                matrix_neutral[0][cod-1]+=1
        else:
            continue
                    
                
            
    #plotting
    if plot==True:
        fig, ax = plt.subplots(5,1,figsize=(25,20), sharex=True, gridspec_kw={ 'height_ratios':[3,8,0.7,.7,.7]})

        g = sns.heatmap(matrix_snp, cmap='Blues', ax=ax[1],xticklabels=10,alpha=1, cbar=False,cbar_kws={"orientation": "horizontal", "pad":0.05, "label":'Number of Occurences'})

        
        cmap_name='cyan1'
        colors = ['white','Grey']
        cmap = LinearSegmentedColormap.from_list(cmap_name, colors, N=4)
        h = sns.heatmap(matrix_neutral, cmap=cmap, ax=ax[2],xticklabels=10,alpha=1,cbar=False,cbar_kws={"orientation": "horizontal", "pad":0.05, "label":'Number of Occurences'})
        
        f = sns.heatmap(matrix_ins, cmap='Greens', vmin=0,ax=ax[3],xticklabels=10,alpha=1,cbar=False, cbar_kws={"orientation": "horizontal", "pad":0.05, "label":'Number of Occurences'})

        q = sns.heatmap(matrix_del, cmap='Purples', ax=ax[4],xticklabels=10,alpha=1,cbar=False,cbar_kws={"orientation": "horizontal", "pad":0.05, "label":'Number of Occurences'})
        
        zero = np.zeros(393)+9
        zero[0:41]=1 #AD1
        zero[42:91]=2 #AD2
        zero[101:291]=3 #DBD
        zero[304:321]=4 #NLS
        zero[325:355]=5 #TD
        zero[363:392]=6 #
        #p = sns.heatmap([zero], cmap='Pastel1', ax=ax[0], cbar=False, yticklabels = False,xticklabels = 20)#,cbar_kws = dict(use_gridspec=True,cax=ax[1][2]))
        
        #p.tick_params(bottom=False) 


        f.tick_params(bottom=False) 
        g.tick_params(bottom=False) 

        h.tick_params(bottom=False) 


                
        ax[0].bar(range(1, len_gene),zero_array[0], color='tab:blue',alpha=0.7, label='SNP')
        ax[0].bar(range(1, len_gene),zero_array[1], bottom=zero_array[0], color='tab:green',alpha=0.7, label='INS')
        ax[0].bar(range(1, len_gene),zero_array[2], bottom=zero_array[0]+zero_array[1], color='tab:purple', alpha=0.7,label='DEL')
        ax[0].bar(range(1, len_gene),zero_array[3],bottom=zero_array[0]+zero_array[1]+zero_array[2], color='tab:grey',alpha=0.7, label='ONP (neutral)')


        
        ax[0].set_xlim(1,len_gene)
        ax[0].set_ylabel('Number of pegRNAs', fontsize=17) 
        ax[4].set_xlabel('Codon', fontsize=15)
        ax[1].set_yticklabels(amino_acids, fontsize=15, rotation=0)
        ax[2].set_yticklabels(['NEUTRAL'], fontsize=15, rotation=0)
        ax[3].set_yticklabels(['INS'], fontsize=15, rotation=0)

        ax[4].set_yticklabels(['DEL'], fontsize=15, rotation=0)



        
        ax[0].set_title(gene_name, fontsize=25)

        for i in range(23):
            ax[1].axhline(i, color='white', lw=2)

        #h = sns.heatmap([zero_array], cmap='Reds', ax=ax[0])
        plt.subplots_adjust(wspace=0, hspace=0.05)
    
    #fig.colorbar(g, orientation="horizontal")


    
        return matrix_snp,matrix_ins, matrix_del, matrix_neutral, fig
    
    return matrix_snp,matrix_ins, matrix_del, matrix_neutral


# main functions: calcAllScores and calcMutSeqs 

# this library re-implements the efficiency scoring functions of these articles in calcAllScores():

# - WangSvm: Wang et al, Science 2014, PMID 24336569, no website
# - Doench: Doench et al, Nat Biotech 2014, PMID 25184501, http://www.broadinstitute.org/rnai/public/analysis-tools/sgrna-design
# - CrisprScan: Moreno-Mateos, Nat Meth 2015, PMID 26322839, http://crisprscan.org
# - ssc: Xu et al, Gen Res 2015, PMID 26063738, http://crispr.dfci.harvard.edu/SSC/
# - Chari: Chari et al, PMID 26167643 http://crispr.med.harvard.edu/sgRNAScorer
# - Fusi: Fusi et al, prepublication manuscript on bioarxiv, http://dx.doi.org/10.1101/021568 http://research.microsoft.com/en-us/projects/azimuth/, only available as a web API
# - Housden: Housden et al, PMID 26350902, http://www.flyrnai.org/evaluateCrispr/
# - Wu-Crispr: Wong et al, http://www.genomebiology.com/2015/16/1/218
# - DeepCpf1, Kim et al, PMID 29431740, https://www.ncbi.nlm.nih.gov/pubmed/29431740
# - SaCas9 efficiency score (no name), Najm et al, https://www.ncbi.nlm.nih.gov/pubmed/29251726

# Also includes the prediction of DSB-repair outcome in calcMutSeqs:
# - OOF: Microhomology and out-of-frame score from Bae et al, Nat Biotech 2014 , PMID24972169 http://www.rgenome.net/mich-calculator/
# - Wei Chen et al: 

# the input are 100bp sequences that flank the basepair just 5' of the PAM +/-50bp.
# so 50bp 5' of the PAM, and 47bp 3' of the PAM -> 100bp

# this module uses pipes to feed data into some programs
# If you run too many sequences at once, it may hang. Increase the BUFSIZE variable in this case.

from subprocess import Popen, PIPE, STDOUT, check_output, CalledProcessError, call
import platform, math, tempfile, bisect, sys, os, logging, types, optparse, shutil
from os.path import dirname, join, basename, isfile, expanduser, isdir, abspath
from math import log10

import urllib.request, urllib.error, urllib.parse, pickle
import json

myDir = dirname(__file__)

aziDir = join(myDir, "bin/Azimuth-2.0/")
sys.path.append(aziDir)

global binDir
binDir = None

# the name of a directory to use for caching some efficiency values that are slow to calculate
# deactivated by default
cacheDir = None

# by default bindir is relative to the location of this library
if binDir is None:
    binDir = join(dirname(__file__), "bin")

BUFSIZE = 10000000

def setBinDir(path):
    global binDir
    binDir = path

def setCacheDir(path):
    global cacheDir
    cacheDir = path

def getBinPath(name, isDir=False):
    """
    get the full pathname of a platform-specific binary, in the bin/ directory relative to this directory
    """
    currPlatform = platform.system()
    binPath = join(binDir, currPlatform, name)
    if isDir and not isdir(binPath):
        raise Exception("Could not find directory %s" % binPath)
    if not isDir and not isfile(binPath):
        raise Exception("Could not find file %s" % binPath)
    return binPath

def seqToVec(seq, offsets={"A":0,"C":1,"G":2,"T":3}):
    """ convert a x bp sequence to a 4 * x 0/1 vector
    >>> seqToVec("AAAAATTTTTGGGGGCCCCC")
    [1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
    """
    assert(len(seq)==20)
    row = [0]*len(seq)*4
    pseudoOffset = offsets["A"]
    for pos, nucl in enumerate(seq):
        nucl = nucl.upper()
        # treat N, Y, etc like "A". Happens very rarely.
        nuclOffset = offsets.get(nucl, pseudoOffset)
        vecPos = (pos*len(offsets))+nuclOffset
        #if vecPos not in range(len(row)):
            #ofh = open("temp.txt", "a")
            #ofh.write(str(vecPos)+" "+seq+" "+str(row)+"pos %d, nucl %s" % (pos, nucl)+"\n")
            #assert(False)
        row[vecPos] = 1
    return row

def vecToSeqDicts(coefs):
    " convert a list of 80 floats to 20 dictionaries with A/C/T/G -> float "
    freqs = []
    for i in range(0,20):
        charFreqs = {}
        for nucl, x in zip("ACGT", list(range(0,4))):
            freq = coefs[i*4+x]
            if freq==0.0:
                continue
            charFreqs[nucl] = freq
        freqs.append(charFreqs)
    return freqs


def writeDict(d, fname):
    " write dict as a tab file "
    if not isdir(dirname(fname)):
        logging.debug("Cannot write %s, no caching of efficiency scores" % fname)
        return

    ofh = open(fname, "w")
    for k, v in d.items():
        if type(v)==tuple:
            ofh.write("%s\t%s\n" % (k, "\t".join([str(x) for x in v])))
        else:
            ofh.write("%s\t%s\n" % (k, str(v)))
    ofh.close()

def readDict(fname, isFloat=True):
    " read dict from a tab sep file "
    if not isfile(fname):
        logging.debug("%s does not exist. Returning empty dict" % fname)
        return {}

    logging.info("Reading %s" %fname)
    data = {}
    for line in open(fname):
        fs = line.rstrip("\n").split("\t")
        if len(fs)==2:
            k, v = fs
            if isFloat:
                v = float(v)
        else:
            k = fs[0]
            v = tuple(fs[1:])
            if isFloat:
                v = tuple([float(x) for x in v])
        data[k] = v
    return data

class ScoreCache:
    """
    a cache of eff scores, kept on disk. Can avoid slow calculations by keeping
    the value of the score in a tab-sep file.
    """

    def __init__(self, scoreName):
        self.cacheFname = join(cacheDir, "%s.tab" % scoreName)
        scoreCache = readDict(self.cacheFname, isFloat=True)
        self.scoreCache = scoreCache

    def findNewSeqs(self, seqs):
        """ get seqs that are not in cache. If all are, return the list of scores.
        Otherwise return None for the scores.
        Returns tuple (seqs, scores)
        """
        self.allSeqs = seqs
        newSeqs = set()
        for s in seqs:
            if not s in self.scoreCache:
                newSeqs.add(s)

        scoreList = None
        if len(newSeqs)==0:
            scoreList = [self.scoreCache[s] for s in seqs]
        self.newSeqs = newSeqs
        return newSeqs, scoreList

    def mergeIntoCache(self, newScores):
        # create final result merging cache and newly obtained scores
        scoreList = []
        assert(len(newScores)==len(self.newSeqs))
        newScoreDict = dict(list(zip(self.newSeqs, newScores)))

        for s in self.allSeqs:
            if s in newScoreDict:
                scoreList.append(newScoreDict[s])
            else:
                scoreList.append(self.scoreCache[s])

        self.scoreCache.update(newScoreDict)
        writeDict(self.scoreCache, self.cacheFname)
        return scoreList

def calcAziScore(seqs):
    " the official implementation of the Doench2016 (aka Fusi) score from Microsoft "
    import numpy
    import azimuth.model_comparison
    res = []
    for seq in seqs:
        if "N" in seq:
            res.append(-1) # can't do Ns
            continue

        pam = seq[25:27]
        # pam_audit = do not check for NGG PAM
        seq = seq.upper()
        score = azimuth.model_comparison.predict(numpy.array([seq]), None, None, pam_audit=False)[0]
        res.append(int(round(100*score)))
    return res

# ----------- MAIN --------------
if __name__=="__main__":
    import cProfile
    #cProfile.runctx('print runLindel(["test"], ["CCCTGGCGGCCTAAGGACTCGGCGCGCCGGAAGTGGCCAGGGCGGGGGCGACCTCGGCTCACAG"])', globals(), locals())
    #pr = cProfile.Profile()
    #pr.enable()
    #runLindel(["test"], ["CCCTGGCGGCCTAAGGACTCGGCGCGCCGGAAGTGGCCAGGGCGGGGGCGACCTCGGCTCACAG"])
    #pr.disable()
    #pr.print_stats(sort='tottime')
    #sys.exit(0)


    args, options = parseArgs()
    if options.test:
        test()
        sys.exit(0)

    #setBinDir("../crispor/bin")
    setBinDir("./bin")
    inFname = sys.argv[1]
    seqs = readSeqs(inFname)
    if len(seqs)==0:
        logging.error("No sequences in input left")
    else:
        printScoreTabSep(seqs, options.all, options.enzyme)



import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt 
plt.rc('font', family='Helvetica')
import Bio.Seq
import gzip
from Bio import SeqIO
from Bio.Align import PairwiseAligner
from Bio.pairwise2 import format_alignment
import warnings
import regex as re
import pickle
from matplotlib.colors import LinearSegmentedColormap
import seaborn as sns
import matplotlib.patches as patches
from matplotlib.patches import Polygon
from cyvcf2 import VCF
import crisporEffScores #python script included in directory; from CRISPOR website github with modifications

warnings.filterwarnings('ignore')

#--- GLOBAL VARIABLES ---------
MATCH_SCORE = 1
MISMATCH_SCORE = -0.5
OPEN_GAP_SCORE = -5
EXTEND_GAP_SCORE = -0.1
TARGET_END_GAP_SCORE = 0
QUERY_END_GAP_SCORE = 0

def make_aligner():
    """ 
    Aligner function from Bio.Align.PairwiseAligner with custom parameters.
    """
    aligner = PairwiseAligner()
    aligner.mode = 'global'
    aligner.match_score = MATCH_SCORE
    aligner.mismatch_score = MISMATCH_SCORE
    aligner.open_gap_score = OPEN_GAP_SCORE
    aligner.extend_gap_score = EXTEND_GAP_SCORE
    aligner.target_end_gap_score = TARGET_END_GAP_SCORE
    aligner.query_end_gap_score = QUERY_END_GAP_SCORE
    return aligner

aligner = make_aligner()

class mutation:
    """ 
    Class for storing information about individual mutations. Used by functions throughout.
    Prevents the use of disorganized lists.
    """
    def __init__(self, wt_w_context, alt_w_context, left_seq, right_seq, var_type, ref_seq, alt_seq, chrom=None, genome=None):
        """ 
        change seq_start and seq_end to correspond with chromosome coordinates
        or some other organizational method...
        """
        
        #WT sequence
        self.wt_forward = wt_w_context.upper()
        rc = str(Bio.Seq.Seq(self.wt_forward).reverse_complement())
        self.wt_rc = rc

        #ALT sequence
        self.mut_forward = alt_w_context.upper()
        rc_mut = str(Bio.Seq.Seq(self.mut_forward).reverse_complement())
        self.mut_rc = rc_mut

        #seq start refers to the WIDE sequence start (i.e. the WT sequence with e.g. 60 nt of context)
        self.seq_start = 0
        self.seq_end = len(self.wt_forward)
        self.chrom = chrom
        self.genome_build = genome

        #and splitting it up into the constituent components
        self.left_seq = left_seq.upper()
        self.right_seq = right_seq.upper()
        self.ref_seq = ref_seq.upper()
        self.alt_seq = alt_seq.upper()

        #and do the same to get it into the RC format

        #RIGHT/LEFT ARE SWAPPED ON REVERSE COMPLEMENT
        self.left_seq_rc = str(Bio.Seq.Seq(right_seq).reverse_complement())
        self.right_seq_rc = str(Bio.Seq.Seq(left_seq).reverse_complement())
        self.ref_seq_rc = str(Bio.Seq.Seq(ref_seq).reverse_complement())
        self.alt_seq_rc = str(Bio.Seq.Seq(alt_seq).reverse_complement())

        self.variant_type = var_type
        #self.alt_size = (positive or negative integer)
        
        #information about the PAM sequence location
        self.PAM_idx_forward = None
        self.PAM_idx_rc = None


#-----------functions------------
def genome_loader(filepath_gz):
    """
    Takes in filepath of human or mouse genome and returns dictionary of chromosome sequences that PEGG can parse.
    Tested only on human and mouse genomes.

    Returns (1) chromosome dictionary, and (2) file names of the chromosomes that are stored (for manual checking of errors).
    
    Parameters
    -----------
    filepath_gz
        *type = str*
        
        The filepath to the .gz file holding the reference genome file.
 
    """
    #------loading in reference genome and organizing it into a 2-d list by chromosome---------------------
    wrong = ["alternate", "unplaced", "unlocalized", "patch", "mitochondrion"]
    filtered = []
    chrom_list = []
    seq_list = []

    with gzip.open(filepath_gz, "rt") as handle:
        #records = list(SeqIO.parse(handle, "fasta")) #about 4 Gb in  memory

        for i in SeqIO.parse(handle, "fasta"):
            ii = i.description
            ignore=False 

            for key in wrong:
                if key in ii:
                    ignore = True
            
            if ignore==False:

                x = ii.find('chromosome')
                chrom = ii[x:].split(' ')[1]
                chrom = chrom.split(',')[0]
                if chrom not in ['X', 'Y']:
                    chrom = int(chrom)
                chrom_list.append(chrom)
                seq_list.append(i.seq)

                filtered.append(ii)

  
    return dict(zip(chrom_list, seq_list)), filtered

def df_formatter(df, chrom_dict, context_size = 120): 
                 #cols_to_save = ['Hugo_Symbol', 'Chromosome', 'Start_Position', 'End_Position', 'Consequence', 'Variant_Classification', 'Variant_Type', 'Reference_Allele', 'Tumor_Seq_Allele2',  'HGVSc', 'HGVSp_Short']):

    """ 
    Takes in variants (in cBioPortal format!) and outputs dataframe with REF and ALT oligos with designated context_size
    that can be used by PEGG for creating pegRNAs.

    Parameters
    -----------
    df
        *type = pd.DataFrame*
        
        The dataframe of input variants in cBioPortal format.

    chrom_dict
        *type = dict*

        Dictionary generated by genome_loader() that holds the chromosome sequences.

    context_size
        *type = int*

        The amount of context/flanking sequence on either side of the mutation to generate. For larger variants, set this larger.
        Default = 120. e.g. AAA(A/G)AAA = context_size of 3
 
    """

    wt_w_context = []
    alt_w_context = []
    left_context_list = []
    right_context_list = []
    ref_allele = []
    alt_allele = []

    seq_start = []
    seq_end = []

    to_drop = []

    for i, val in df.iterrows():
        vt = val['Variant_Type']
        s = val['Start_Position']
        e = val['End_Position']
        ref = val['Reference_Allele']
        alt = val['Tumor_Seq_Allele2']
        chrom = val['Chromosome']

        if chrom not in ['X','Y']:
            chrom = int(chrom)
       
        chr_seq = chrom_dict[chrom].upper()

        if vt in ['SNP', 'ONP', 'DNP', 'INDEL']:
            ref = ref
            alt = alt
            #assert ref == chr_seq[s-1:e], print(ref, chr_seq[s-1:e])
            left_context = chr_seq[s-1-context_size:s-1]
            right_context = chr_seq[e:e+context_size]

        elif vt =='INS':
            ref = ''
            alt = alt
            #left_context = chr_seq[s-1-context_size:s+1] #need to do this since INS reference alleles are blank
            left_context = chr_seq[s-1-context_size:s]
            right_context = chr_seq[e-1:e+context_size]

        elif vt=='DEL':
            ref = ref
            alt = ''
            left_context = chr_seq[s-1-context_size:s-1]
            right_context = chr_seq[e:e+context_size]

        wt_seq = left_context + ref + right_context
        alt_seq = left_context + alt + right_context

        

        start = s-context_size
        end = e+context_size

        if str(chr_seq[start-1:end])!=str(wt_seq): #changed from assert to allow it to continue to run

            print(f"Error in mutant # {i}; Dropped from pegRNA generation\nVariant Type = {vt}| REF = {ref} | ALT = {alt}\n{chr_seq[start-1:end]}\n{str(wt_seq)}")
            to_drop.append(i)
        else:
            seq_start.append(start)
            seq_end.append(end)
            
            wt_w_context.append(str(wt_seq))
            alt_w_context.append(str(alt_seq))
            left_context_list.append(str(left_context))
            right_context_list.append(str(right_context))
            ref_allele.append(str(ref))
            alt_allele.append(str(alt))

    cols_to_save = list(df.keys())

    df_new = df[cols_to_save]
    df_new = df_new.drop(index=to_drop).reset_index(drop=True)

    df_new['seq_start'] = seq_start
    df_new['seq_end'] = seq_end
    df_new['wt_w_context'] = wt_w_context
    df_new['alt_w_context'] = alt_w_context
    df_new['left_context'] = left_context_list
    df_new['right_context'] = right_context_list
    df_new['REF'] = ref_allele
    df_new['ALT'] = alt_allele
    df_new = df_new.reset_index()
    
    return df_new

def mut_formatter(wt, alt):
    """ 
    Formats mutations for pegRNA generation. Takes in WT and ALT sequences, returns necessary parameters for gRNA/pegRNA generation.
    Note: This function will break with INDELs and complex variants since it relies on a simple alignment.

    Parameters
    -----------
    wt
        *type = str*
        
        WT sequence

    alt
        *type = str*

        Mutant/alternate sequence.
    """

    alignments = aligner.align(wt, alt)

    WT_align = alignments[0].aligned[0]
    alt_align = alignments[0].aligned[1]

    if len(WT_align) == 1:
        #Substitution (SNP or ONP)

        #find indexes of non-matching values to get right/left context sequences and mutant sequence
        non_matching_idx = []

        for i, val in enumerate(list(zip(wt, alt))):
            if val[0] != val[1]:
                non_matching_idx.append(i)
        
        assert len(non_matching_idx) !=0, f'No mutation detected in alt sequence provided | row index = {i}'

        if len(non_matching_idx)==1:
            var_type = 'SNP'
        else:
            var_type = 'ONP'

        left_seq = wt[:non_matching_idx[0]]
        ref_seq = wt[non_matching_idx[0]:non_matching_idx[-1]+1]
        alt_seq = alt[non_matching_idx[0]:non_matching_idx[-1]+1]
        right_seq = wt[non_matching_idx[-1]+1:]


    elif len(WT_align)>1:
        if len(alt)>len(wt):
            var_type = 'INS'
        elif len(alt)<len(wt):
            var_type='DEL'

        left_seq = wt[WT_align[0][0]:WT_align[0][1]]
        right_seq = wt[WT_align[1][0]:WT_align[1][1]]

        if var_type=='INS':
        #INS or DEL
            ref_seq = ''
            alt_seq = alt[alt_align[0][1]:alt_align[1][0]]
            
        elif var_type =='DEL':
            ref_seq = wt[WT_align[0][1]:WT_align[1][0]]
            alt_seq = ''
            

    return var_type, left_seq, ref_seq, alt_seq, right_seq


def primedesign_formatter(seq):
     
    """ 
    Takes as input sequence in prime design format e.g. AATTCCG(G/C)AATTCGCT. Returns necessary parameters for pegRNA generation.

    Parameters
    -----------
    seq
        *type = str*
        
        The PrimeDesign formatted sequence.
    """

    start = seq.find("(")

    end = seq.find(")")

    replace_seq = seq[start:end+1]
    #if '/' in replace_seq:
        #if '+' in replace_seq:
            #throw an error

    loc_replace = replace_seq.find('/')
    ref_seq = replace_seq[1:loc_replace]
    alt_seq = replace_seq[loc_replace+1:-1]

    left_seq = seq[:start]
    right_seq = seq[end+1:]

    wt_w_context = left_seq + ref_seq + right_seq
    alt_w_context = left_seq + alt_seq + right_seq

    if len(ref_seq)==0:
        var_type='INS'

    else:
        if len(ref_seq)==len(alt_seq):
            if len(ref_seq)==1:
                var_type='SNP'
            else:
                var_type='ONP'

        else:
            if len(alt_seq) == 0:
                var_type = 'DEL'
            else:
                var_type = 'INDEL'

    return var_type, wt_w_context, alt_w_context, left_seq, right_seq, ref_seq, alt_seq


def clinvar_VCF_translator(filepath, variation_ids):
    """
    Function that takes a clinvar.vcf.gz file containing information about variants, as well as a list of variation ID numbers,
    and returns a pandas dataframe containing the variants in a format that can be used by PEGG to design pegRNAs.
    
    Parameters
    ----------
    filepath
        *type = str*
        
        Filepath to the clinvar.vcf.gz file.
        
    variation_ids
        *type = list*
        
        List of variation IDs that the user wants to convert to .
        
    """
    gene = []
    ref = []
    alt = []
    chrom = []
    start = []
    end = []
    CLNHGVS = []
    var_type = []
    CLNDN = []
    CLNSIG = []
    allele_id = []
    
    var_id = []

    #for translating between Clinvar variation classes, and PEGG-readable versions...
    a = ['Deletion', 'Duplication', 'Indel', 'Insertion', 'Inversion',
        'Microsatellite', 'Variation', 'single_nucleotide_variant']
    b = ['DEL', 'Duplication', 'INDEL', 'INS', 'Inversion',
        'Microsatellite', 'Variation', 'SNP']

    var_dict = dict(zip(a,b))


    for variant in VCF(filepath): # or VCF('some.bcf')
        if int(variant.ID) in variation_ids:
            
            var_id.append(int(variant.ID))
            allele_id.append(variant.INFO.get('ALLELEID'))
            #change it to make it flexible for a list
            gene.append(variant.INFO.get('GENEINFO'))
            CLNSIG.append(variant.INFO.get('CLNSIG'))
            var_type.append(variant.INFO.get('CLNVC'))
            CLNHGVS.append(variant.INFO.get('CLNHGVS'))
            CLNDN.append(variant.INFO.get('CLNDN'))
            ref.append(variant.REF)
            chrom.append(variant.CHROM)
            start.append(variant.start)
            end.append(variant.end)

            if len(variant.ALT)==1:
                alt.append(variant.ALT[0])
            elif len(variant.ALT)>1:
                tot = len(variant.ALT)-1
                alt.append(variant.ALT[0])
                for i in range(tot):
                    var_id.append(int(variant.ID))
                    allele_id.append(variant.INFO.get('ALLELEID'))
                    #change it to make it flexible for a list
                    gene.append(variant.INFO.get('GENEINFO'))
                    CLNSIG.append(variant.INFO.get('CLNSIG'))
                    var_type.append(variant.INFO.get('CLNVC'))
                    CLNHGVS.append(variant.INFO.get('CLNHGVS'))
                    CLNDN.append(variant.INFO.get('CLNDN'))
                    ref.append(variant.REF)
                    chrom.append(variant.CHROM)
                    start.append(variant.start)
                    end.append(variant.end)
                    alt.append(variant.ALT[i+1])
                    #d1.append(variant.ID)

    var_type = [var_dict[i] for i in var_type]
    gene = [i.split(':')[0] for i in gene] #extracting just gene information...

    #fixing issue where Clinvar variants start one position early for some reason...
    start = np.asarray(start)+1
    
    d1 = ['Hugo_Symbol', 'Chromosome', 'Start_Position', 'End_Position', 'Reference_Allele', 
          'Tumor_Seq_Allele2', 'Variant_Type', 'Variation_ID','Allele_ID','CLNSIG', 'CLNHGVS', 'CLNDN']

    combined = [gene, chrom, start, end, ref, alt, var_type, var_id, allele_id,
               CLNSIG, CLNHGVS, CLNDN]
    
    d = dict(zip(d1, combined))
    clinvar = pd.DataFrame(data = d)

    return clinvar

def PAM_finder(seq, PAM):
    """ 
    Finds indeces of PAM sequences in sequence.

    Parameters
    ------------
    seq
        *type = str*

        Sequence to search for PAM sequences.

    PAM
        *type = str*
        
        PAM sequence for searching. Formatting: e.g. "NGG"

        N = [A|T|C|G]
        R = [A|G]
        Y = [C|T]
        S = [G|C]
        W = [A|T]
        K = [G|T]
        M = [A|C]
        B = [C|G|T]
        D = [A|G|T]
        H = [A|C|T]
        V = [A|C|G]
    """
    PAM_dict = {"A":"A",
                "G":"G",
                "C":"C",
                "T":"T",
                "N":"[A|T|C|G]",
                "R":"[A|G]",
                "Y":"[C|T]",
                "S":"[G|C]",
                "W":"[A|T]",
                "K":"[G|T]",
                "M":"[A|C]",
                "B":"[C|G|T]",
                "D":"[A|G|T]",
                "H":"[A|C|T]",
                "V":"[A|C|G]"}

    PAM_new = ""
    for i in PAM:
        PAM_new += PAM_dict[i]

    p = re.compile(PAM_new)

    iterator = p.finditer(seq, overlapped=True)
    PAM_holder = []
    for match in iterator:
        PAM_holder.append(match.span())

    return PAM_holder


def eligible_PAM_finder(mut, PAM, max_RTT_length, proto_size=19):
    """ 
    Determines eligible PAM sequences for creating pegRNAs.
    Returns PAM sequence locations on (1) forward, and (2) reverse-complement strand

    Parameters
    ------------
    mut
        *type = mutation class*

        See class: mutation

    PAM
        *type = str*
        
        PAM sequence for searching. Formatting: e.g. "NGG"

        N = [A|T|C|G]
        R = [A|G]
        Y = [C|T]
        S = [G|C]
        W = [A|T]
        K = [G|T]
        M = [A|C]
        B = [C|G|T]
        D = [A|G|T]
        H = [A|C|T]
        V = [A|C|G]

    max_RTT_length
        *type = int*

        Max RTT length for pegRNAs being designed.

    proto_size
        *type = int*

        Size of protospacer being used. Default = 19 (G+19).
    """

    forward = PAM_finder(mut.wt_forward, PAM)
    reverse = PAM_finder(mut.wt_rc, PAM)

    var_type = mut.variant_type

    #first do the forward sequence
    left_len_F = len(mut.left_seq)
    left_len_R = len(mut.left_seq_rc)

    if var_type =='INS':
        alt_size = len(mut.alt_seq)
        m_start_F = left_len_F - (max_RTT_length - alt_size - 3)
        m_end_F = left_len_F + 4
        
        m_start_R = left_len_R - (max_RTT_length - alt_size - 3)
        m_end_R = left_len_R + 4

    else:
        alt_size = len(mut.alt_seq)
        m_start_F = left_len_F - (max_RTT_length - alt_size - 3)
        m_end_F = left_len_F + 3

        m_start_R = left_len_R - (max_RTT_length - alt_size - 3)
        m_end_R = left_len_R + 3


    eligible_PAM_start_F = max(m_start_F, proto_size + 3) #
    eligible_PAM_end_F = m_end_F
    
    eligible_PAM_start_R = max(m_start_R, proto_size + 3) #
    eligible_PAM_end_R = m_end_R

    eligible_PAMS_F = []
    eligible_PAMS_R = []

    for i in forward:
        #check start of PAM site
        s = i[0]
        if (s>=eligible_PAM_start_F) and (s<=eligible_PAM_end_F):
            eligible_PAMS_F.append(i)

    for i in reverse:
        #check start of PAM site
        s = i[0]
        if (s>=eligible_PAM_start_R) and (s<=eligible_PAM_end_R):
            eligible_PAMS_R.append(i)

    return eligible_PAMS_F, eligible_PAMS_R

def pegRNA_generator(mut, PAM, orientation, proto_size, RTT_lengths, PBS_lengths):
    """ 
    Generates pegRNAs for a given mutation.

    Parameters
    ------------
    mut
        *type = mutation class*

        See class: mutation

    PAM
        *type = str*
        
        PAM sequence for searching. Formatting: e.g. "NGG"

        N = [A|T|C|G]
        R = [A|G]
        Y = [C|T]
        S = [G|C]
        W = [A|T]
        K = [G|T]
        M = [A|C]
        B = [C|G|T]
        D = [A|G|T]
        H = [A|C|T]
        V = [A|C|G]

    orientation
        *type = str*

        '+' or '-'

    proto_size
        *type = int*

        Size of protospacer being used. Default = 19 (G+19).

    RTT_lengths
        *type = list*

        List containing RTT lengths to design pegRNAs for.

    PBS_length
        *type = list*

        List containing PBS lengths to desing pegRNAs for.
    """


    #------function----
    protospacer_seqs = []
    protospacer_wide_seqs = []
    PAM_start_list = []
    PAM_end_list = []
    PAM_list = []

    RTT_seqs = []
    RTT_size = []
    PBS_seqs = []
    PBS_size = []
    RTT_PBS_seqs = []

    #descriptors of pegRNAs...
    distance_to_nick = [] #distance from the start of the mutation to the nick introduced by nCas9 (-3 from PAM)
    RHA = [] #right homology arm (distance from end of mutation to end of RTT)
    PAM_disrupted_list = []
    proto_disrupted_list = []

    #Get mutation info, depending on whether we're looking on the + or the - strand
    if orientation == '+':
        left_len = len(mut.left_seq)
        ref_len = len(mut.ref_seq)
        alt_len = len(mut.alt_seq)
        alt_seq = mut.alt_seq
        seq_F = mut.wt_forward

        PAM_seqs = mut.PAM_idx_forward

    elif orientation == '-':
        left_len = len(mut.left_seq_rc)
        ref_len = len(mut.ref_seq_rc)
        alt_len = len(mut.alt_seq_rc)
        alt_seq = mut.alt_seq_rc
        seq_F = mut.wt_rc

        PAM_seqs = mut.PAM_idx_rc

    #iterate over the PAM sequences:
    for i in PAM_seqs:
        PAM_start = i[0]
        PAM_end = i[1]

        RTT_start = PAM_start - 3

        left_RTT = seq_F[RTT_start:left_len]

        for RTT_length in RTT_lengths:

            remaining_length = RTT_length - (len(left_RTT) + alt_len)

            #make sure RTT is long enough
            #and make sure that there's sufficient sequence to actually pull from...
            if (remaining_length <0) or ((left_len+ref_len + remaining_length) > len(seq_F)):
                RTT_seqs.append(None)
                distance_to_nick.append(None)
                RHA.append(None)
                PAM_disrupted_list.append(None)
                proto_disrupted_list.append(None)
                PAM_start_list.append(None)
                PAM_end_list.append(None)
                PAM_list.append(None)
                protospacer_seqs.append(None)
                protospacer_wide_seqs.append(None)
                PBS_seqs.append(None)
                RTT_PBS_seqs.append(None)
                RTT_size.append(None)
                PBS_size.append(None)

            else: #able to design the pegRNA

                right_RTT = seq_F[left_len+ref_len:left_len+ref_len + remaining_length]
                RTT = left_RTT + alt_seq + right_RTT

                #and then determine if the PAM sequence is disrupted
                #protospacer disrupted would just be if the distance to nick is < 3
                #fish out the PAM sequence from RTT
                PAM_new = RTT[3:3+len(PAM)]
                pam_list = PAM_finder(PAM_new, PAM)

                PAM_disrupted = False
                if len(pam_list)==0:
                    PAM_disrupted = True

                proto_disrupted = False
                if len(left_RTT) < 3:
                    proto_disrupted = True

                #and finally reverse complement the RTT
                RTT = str(Bio.Seq.Seq(RTT).reverse_complement())

                #pull out the PAM sequence and protospacer as well
                PAM_sequence = seq_F[PAM_start:PAM_end]
                protospacer =  'G' + seq_F[PAM_start - proto_size:PAM_start]

                #extract wide protospacer for Rule set 2 prediction...
                protospacer_wide = seq_F[PAM_start - 20 - 4 : PAM_start + 3 + 3]
                
                #iterate over PBS lengths for adding things to list
                for PBS_length in PBS_lengths:
                    RTT_seqs.append(RTT)
                    RTT_size.append(RTT_length)
                    PBS_size.append(PBS_length)
                    distance_to_nick.append(len(left_RTT))
                    RHA.append(len(right_RTT))
                    PAM_disrupted_list.append(PAM_disrupted)
                    proto_disrupted_list.append(proto_disrupted)

                    #------add in PAM info, protospacer, + PBS/RTT_PBS
                    #record info about PAM sequences
                    if orientation == '+':
                        PAM_start_list.append(PAM_start)
                        PAM_end_list.append(PAM_end)
                    elif orientation == '-':
                        PAM_start_list.append(len(seq_F)-PAM_start)
                        PAM_end_list.append(len(seq_F)-PAM_end)

                    PAM_list.append(PAM_sequence)
                    protospacer_seqs.append(protospacer)
                    protospacer_wide_seqs.append(protospacer_wide)

                    #----and finally design the PBS sequences----
                    #NEED TO HAVE AN ERROR MESSAGE FOR PBS SEQUENCES LARGER THAN 17
                    PBS = seq_F[RTT_start-PBS_length:RTT_start]
                    PBS = str(Bio.Seq.Seq(PBS).reverse_complement())
                    PBS_seqs.append(PBS)

                    #---and then add it to the RTT to make the full 3' extension
                    RTT_PBS_seqs.append(RTT+PBS)

    cols = [PAM_start_list, PAM_end_list, PAM_list, orientation, protospacer_wide_seqs, protospacer_seqs, RTT_seqs, RTT_size, PBS_seqs, PBS_size, RTT_PBS_seqs, distance_to_nick, RHA, PAM_disrupted_list, proto_disrupted_list]
    col_labels = ["PAM_start", "PAM_end", "PAM", "PAM_strand","Protospacer_30", "Protospacer", "RTT", "RTT_length", "PBS", "PBS_length", "RTT_PBS", "Distance_to_nick", "RHA_size", "PAM_disrupted", "Proto_disrupted"]
    dtypes = ['int', 'int', 'str', 'str', 'str', 'str', 'str', 'int', 'str', 'int', 'str', 'int', 'int', 'bool', 'bool']

    dtype_dict = dict(zip(col_labels, dtypes))

    peg_df = pd.DataFrame(dict(zip(col_labels, cols))).dropna().reset_index().drop(columns='index').astype(dtype_dict)

    return peg_df

def sensor_generator(df, proto_size, before_proto_context=5, sensor_length=60, sensor_orientation = 'reverse-complement'):
    """ 
    Generates sensor sequence for quantification of pegRNA editing outcomes.
    Automatically puts sensor in reverse complement orientation with respect to protospacer.
    This is highly reccomended to reduce recombination during cloning/library preparation.

    Parameters
    -----------
    df
        *type = pd.DataFrame*

        Dataframe containing pegRNAs.

    proto_size
        *type = int*

        Size of protospacer. 

    before_proto_context
        *type = int*

        Default = 5. Amount of nucleotide context to put before the protospacer in the sensor

    sensor_length
        *type = int*

        Total length of the sensor in nt. Default = 60. 

    sensor_orientation
        *type = str*

        Options for sensor_orientation = 'reverse-complement' or'forward'.

    """

    assert sensor_orientation in ['reverse-complement', 'forward'], "Choose an eligible sensor orientation option ('reverse_comp' (default) or 'forward')"

    sensor_error = []
    sensor_wt_seqs = []
    sensor_alt_seqs = []

    for i, val in df.iterrows():
        wt = val['wt_w_context']
        alt = val['alt_w_context']
        ref_minus_alt = len(val['REF']) - len(val['ALT'])
        strand = val['PAM_strand']
        pam_start = val['PAM_start']
        pam_end = val['PAM_end']
        RTT_PBS = val['RTT_PBS']
        RHA = val['RHA_size']

        #if PAM is on negative strand; flip it
        if strand == '-':
            pam_start = len(wt) - pam_start
            pam_end = len(wt) - pam_end
            wt = str(Bio.Seq.Seq(wt).reverse_complement())
            alt = str(Bio.Seq.Seq(alt).reverse_complement())

        s1 = pam_start-(proto_size+1)-before_proto_context
        if s1 <0:
            sensor_wt_seqs.append(None)
            sensor_alt_seqs.append(None)
            sensor_error.append('insufficient context sequence provided; try increasing context size')

        else:
            proto_pam = wt[s1:pam_end]
            proto_pam_alt = alt[s1:pam_end]

            remain_len = sensor_length - len(proto_pam)
            if remain_len < 0:
                sensor_wt_seqs.append(None)
                sensor_alt_seqs.append(None)
                sensor_error.append('sensor too small; increase sensor size in parameters; or decrease before_proto_context')

            else:
                remain_len_alt = sensor_length - len(proto_pam) - ref_minus_alt

                e1 = pam_end + remain_len
                e2 = pam_end + remain_len_alt
                if e1>len(wt):
                    sensor_wt_seqs.append(None)
                    sensor_alt_seqs.append(None)
                    sensor_error.append('insufficient context sequence provided; try increasing context size')

                else:
                    rtt_region = wt[pam_end: e1]
                    rtt_region_alt = alt[pam_end:e2]

                    sensor_wt = proto_pam + rtt_region
                    sensor_alt = proto_pam_alt + rtt_region_alt

                    #AND FINALLY either keep it as is, or take the reverse complement
                    if sensor_orientation == 'reverse-complement':
                        sensor_wt = str(Bio.Seq.Seq(sensor_wt).reverse_complement())
                        sensor_alt = str(Bio.Seq.Seq(sensor_alt).reverse_complement())

                        rha_seq = RTT_PBS[:RHA]

                        #check that the homology overhang fits within the sensor sequence
                        if rha_seq not in sensor_wt:
                            sensor_wt_seqs.append(None)
                            sensor_alt_seqs.append(None)
                            sensor_error.append("sensor too small; increase sensor size in parameters; or decrease before_proto_context")

                        else:
                            sensor_wt_seqs.append(sensor_wt)
                            sensor_alt_seqs.append(sensor_alt)
                            sensor_error.append("No Error")
                    
                    if sensor_orientation == 'forward':
                        rha_seq = str(Bio.Seq.Seq(RTT_PBS[:RHA]).reverse_complement())

                        if rha_seq not in sensor_wt:
                            sensor_wt_seqs.append(None)
                            sensor_alt_seqs.append(None)
                            sensor_error.append("sensor too small; increase sensor size in parameters; or decrease before_proto_context")
                    
                        else:
                            sensor_wt_seqs.append(sensor_wt)
                            sensor_alt_seqs.append(sensor_alt)
                            sensor_error.append("No Error")


    df['sensor_wt'] = sensor_wt_seqs
    df['sensor_alt'] = sensor_alt_seqs
    df['sensor_orientation'] = sensor_orientation
    df['sensor_error'] = sensor_error

    return df

def ontarget_score(df):
    """ 
    Calls to crisporEffScores (taken from CRISPOR github) to generate on-target scores using Rule Set 2.
    """

    uniq_protos = list(np.unique(df['Protospacer_30']))

    #make sure they're all the right length
    uniq_protos_true = []
    wrong_length = []
    for val in uniq_protos:
        if len(val)!=30:
            wrong_length.append(val)
        else:
            uniq_protos_true.append(val)

    scores = crisporEffScores.calcAziScore(uniq_protos_true)

    dict_scores = dict(zip(uniq_protos_true, scores))
    dict_none = dict(zip(wrong_length, [None]*len(wrong_length)))

    dict_scores.update(dict_none)

    for protospacer in dict_scores.keys():
        df.loc[df["Protospacer_30"]== protospacer, "OnTarget_Azimuth_Score"] = dict_scores[protospacer]

    #for protospacer in dict_none.keys():
    #    df.loc[df["Protospacer_30"]== protospacer, "OnTarget_Azimuth_Score"] = None

    return df


def input_formatter(input_df, input_format, chrom_dict, context_size):
    """ 
    Master function for putting input mutations in the correct format for pegRNA/gRNA design.

    input_df
        *type = pd.DataFrame*

        Pandas dataframe that contains the input mutations.

    input_format
        *type = str*

        Options = 'cBioPortal', 'WT_ALT', 'PrimeDesign'. For 'WT_ALT', make sure to put headers as "WT" and "ALT". 
        For "PrimeDesign", put the header as "SEQ".

    chrom_dict
        *type = dict*

        Dictionary generated by genome_loader() that holds the chromosome sequences.

    context_size
        *type = int*

        The amount of context/flanking sequence on either side of the mutation to generate. For larger variants, set this larger.
        Default = 120. e.g. AAA(A/G)AAA = context_size of 3

    """

    assert input_format in ['cBioPortal', 'WT_ALT', 'PrimeDesign'], "input_format is not correct. Choose from ['cBioPortal', 'WT_ALT', 'PrimeDesign']."

    #format the input dataframe appropriately...
    if input_format == 'cBioPortal':
        #add assert statement to make sure all necessary info is included
        #MAKE COLS_TO_SAVE MORE FLEXIBLE??? to avoid weird errors...

        input_df = df_formatter(input_df, chrom_dict, context_size)
        input_df['mutation_idx'] = list(range(len(input_df)))

    elif input_format == 'WT_ALT':

        wt = []
        mut = []
        vts = []
        l = []
        ref = []
        alt =[]
        r = []
        for i, val in input_df.iterrows():

            var_type, left_seq, ref_seq, alt_seq, right_seq = mut_formatter(val['WT'], val['ALT'])
            vts.append(var_type)
            l.append(left_seq)
            ref.append(ref_seq)
            alt.append(alt_seq)
            r.append(right_seq)
            wt.append(val['WT'])
            mut.append(val['ALT'])

        mut_idx = list(range(len(wt)))
        cols = [mut_idx, wt, mut, vts, ref, alt, l, r]
        col_labels = ["mutation_idx", 'wt_w_context', 'alt_w_context', 'Variant_Type', 'REF', 'ALT', 'left_context', 'right_context']
        input_df = pd.DataFrame(dict(zip(col_labels, cols)))        

    elif input_format == 'PrimeDesign':

        wt = []
        mut = []
        vts = []
        l = []
        ref = []
        alt =[]
        r = []
        for i, val in input_df.iterrows():
            var_type, wt_w_context, alt_w_context, left_seq, right_seq, ref_seq, alt_seq = primedesign_formatter(val['SEQ'])
            vts.append(var_type)
            l.append(left_seq)
            ref.append(ref_seq)
            alt.append(alt_seq)
            r.append(right_seq)
            wt.append(wt_w_context)
            mut.append(alt_w_context)

        mut_idx = list(range(len(wt)))
        cols = [mut_idx, wt, mut, vts, ref, alt, l, r]
        col_labels = ["mutation_idx", 'wt_w_context', 'alt_w_context', 'Variant_Type', 'REF', 'ALT', 'left_context', 'right_context']
        input_df = pd.DataFrame(dict(zip(col_labels, cols)))     

    return input_df

def other_filtration(pegRNA_df, RE_sites=None, polyT_threshold=4):

    """
    Determines whether pegRNAs contain polyT termination sites or Restriction Enzyme Sites. Returns pegRNA_df with properties added.
    Note: RE_sites requires only A, T, C, or G bases (Ns will not work currently; need to do custom filtration)...
    These properties are checking for:
    
    (1) u6 terminator (polyT sequence)
        
    (2) RE site presence 
        
    Parameters
    ----------
    pegRNA_df
        *type = pd.DataFrame*
        
        A dataframe containing the pegRNAs for the selected input mutations. Generated by run() or pegRNA_generator()

    RE_sites
        *type = list or None*

        A list containing the RE recognition sites to filter (e.g. ['CGTCTC', 'GAATTC'] for Esp3I and EcoRI). Default = None (no filtration).

    polyT_threshold
        *type = int*

        The length of the polyT sequence to classify as a terminator. Default = 4.
    
    """
    #checking the type of dataframe

    prime=False
    base=False
    sensor=False

    df_keys = pegRNA_df.keys()
    if "RTT_PBS" in df_keys:
        prime=True
    else:
        base=True
    if 'sensor_wt' in df_keys:
        sensor=True

    terminator_sequence = 'T'*polyT_threshold

    contains_terminator = []
    RE_site = []

    if RE_sites != None:

        assert type(RE_sites)==list, "RE_sites must be in list format (e.g. ['CGTCTC', 'GAATTC'])"
        
        RE_sites_new = []
        for i in RE_sites:
            for k in i:
                assert k.upper() in ['A', 'T', 'C', 'G'], "RE sites must only contain A, T, C, or G (Ns and other nucleic acid codes not permitted)"
            RE_sites_new.append(i.upper())
            RE_sites_new.append(str(Bio.Seq.Seq(i).reverse_complement()).upper())

        RE_sites = RE_sites_new

        #check for u6 terminator sequence "TTTT" in pbs or protospacer
        for idx, i in pegRNA_df.iterrows():
            #check if its base or prime for filtration purposes            
            if prime==True:
                pbs_rtt = i["RTT_PBS"]
                proto = i["Protospacer"]

                if terminator_sequence in (pbs_rtt or proto):
                    contains_terminator.append(True)
                else:
                    contains_terminator.append(False)

                if sensor==True:
                    target = i['sensor_wt']
                    #finally check for restriction enzyme sites
                    res_proto = any(re_site in proto for re_site in RE_sites)
                    res_pbs_rtt = any(re_site in pbs_rtt for re_site in RE_sites)
                    res_target = any(re_site in target for re_site in RE_sites)

                    if (res_proto or res_pbs_rtt or res_target)==True:
                        RE_site.append(True)
                    else:
                        RE_site.append(False) 
                elif sensor==False:
                    #finally check for restriction enzyme sites
                    res_proto = any(re_site in proto for re_site in RE_sites)
                    res_pbs_rtt = any(re_site in pbs_rtt for re_site in RE_sites)

                    if (res_proto or res_pbs_rtt)==True:
                        RE_site.append(True)
                    else:
                        RE_site.append(False) 
                    

            elif base==True:
                proto = i["Protospacer"]
                if terminator_sequence in (proto):
                    contains_terminator.append(True)
                else:
                    contains_terminator.append(False)

                if sensor==True:
                    target = i['sensor_wt']
                    #finally check for restriction enzyme sites
                    res_proto = any(re_site in proto for re_site in RE_sites)
                    res_target = any(re_site in target for re_site in RE_sites)

                    if (res_proto or res_target)==True:
                        RE_site.append(True)
                    else:
                        RE_site.append(False) 
                elif sensor==False:
                    #finally check for restriction enzyme sites
                    res_proto = any(re_site in proto for re_site in RE_sites)
                    if (res_proto)==True:
                        RE_site.append(True)
                    else:
                        RE_site.append(False) 

        pegRNA_df['contains_polyT_terminator']=contains_terminator
        pegRNA_df['contains_RE_site']=RE_site
    
    if RE_sites == None:
        #check for u6 terminator sequence "TTTT" in pbs or protospacer
        for idx, i in pegRNA_df.iterrows():
            if prime==True:
                pbs_rtt = i["RTT_PBS"]
                proto = i["Protospacer"]

                if terminator_sequence in (pbs_rtt or proto):
                    contains_terminator.append(True)
                else:
                    contains_terminator.append(False)
            elif base==True:
                proto = i["Protospacer"]
                if terminator_sequence in (proto):
                    contains_terminator.append(True)
                else:
                    contains_terminator.append(False)
            
        pegRNA_df['contains_polyT_terminator']=contains_terminator

    return pegRNA_df

def run(input_df, input_format, chrom_dict=None, PAM = "NGG", rankby = 'PEGG2_Score', pegRNAs_per_mut = 'All',
        RTT_lengths = [5,10,15,25,30], PBS_lengths = [8,10,13,15], min_RHA_size = 1,
        RE_sites=None, polyT_threshold=4, 
        proto_size=19, context_size = 120, 
        before_proto_context=5, sensor_length=60, sensor_orientation = 'reverse-complement', sensor=True):

    """ 
    Master function for generating pegRNAs. Takes as input a dataframe containing mutations in one of the acceptable formats.
    Returns a dataframe with pegRNAs with desired design parameters.

    Parameters
    ------------
    input_df
        *type = pd.DataFrame*

        Pandas dataframe that contains the input mutations.

    input_format
        *type = str*

        Options = 'cBioPortal', 'WT_ALT', 'PrimeDesign'. For 'WT_ALT', make sure to put headers as "WT" and "ALT". 
        For "PrimeDesign", put the header as "SEQ".

    chrom_dict
        *type = dict or None*

        Dictionary generated by genome_loader() that holds the chromosome sequences.

    PAM
        *type = str*

        PAM sequence for searching. Default = "NGG". Can include any nucleic acid code (e.g. PAM = "NRCH").

    rank_by
        *type = str*

        What pegRNA parameter to rank pegRNAs by. Options = "PEGG2_Score" (weighted linear regression of different pegRNA parameters)
        or "RF_Score" (random forest predictor of pegRNA efficiency).
    
    pegRNAs_per_mut
        *type = 'All' or int*

        How many pegRNAs to produce per mutation. Default = 'All' (all possible pegRNAs with parameters). Otherwise, choose an integer value (e.g. 5).
    
    RTT_lengths
        *type = list*

        List containing RTT lengths to design pegRNAs for.

    PBS_length
        *type = list*

        List containing PBS lengths to desing pegRNAs for.

    min_RHA_size
        *type = int*

        Minimum size of the RHA (Right homology arm). Default = 1. Generally pegRNAs with smaller RHA perform poorly.

    RE_sites
        *type = list or None*

        A list containing the RE recognition sites to filter (e.g. ['CGTCTC', 'GAATTC'] for Esp3I and EcoRI). Default = None (no filtration).

    polyT_threshold
        *type = int*

        The length of the polyT sequence to classify as a terminator. Default = 4.

    proto_size
        *type = int*

        The length of the protospacer (excluding the appended G at the 5' end). Default = 19 (G+19).

    context_size
        *type = int*

        The amount of context/flanking sequence on either side of the mutation to generate. For larger variants, set this larger.
        Default = 120. e.g. AAA(A/G)AAA = context_size of 3

    before_proto_context
        *type = int*

        Default = 5. Amount of nucleotide context to put before the protospacer in the sensor

    sensor_length
        *type = int*

        Total length of the sensor in nt. Default = 60. 

    sensor_orientation
        *type = str*

        Options for sensor_orientation = 'reverse-complement' or'forward'.

    sensor
        *type = bool*

        True/False whether to include a sensor in the pegRNA design or not. 
    """


    if PAM != 'NGG':
        print('Warning: Efficiency predictions on non-NGG PAMs are un-tested')

    for i in PBS_lengths:
        assert i<=17, "Max PBS Length = 17; rerun with smaller PBS length (or use preset parameters)"

    for i in RTT_lengths:
        if i>30:
            print("Warning: RTT lengths larger than 30 nt are not reccomended due to potential low PE efficiency")
            
    ##enforce the proper formatting for INS, DEL, and INDEL
    #Note: for some reason, Clinvar encodes INS/DEL differently than cBioPortal so most INS/DEL need to be labelled as INDELs
    if input_format == 'cBioPortal':
        input_df.loc[((input_df['Variant_Type']=='INS') & (~input_df['Reference_Allele'].isin(['-', '', None]))), 'Variant_Type'] = 'INDEL'
        input_df.loc[((input_df['Variant_Type']=='DEL') & (~input_df['Tumor_Seq_Allele2'].isin(['-', '', None]))), 'Variant_Type'] = 'INDEL'

        assert chrom_dict != None, "Genome required to use cBioPortal input format. Load genome with genome_loader() and re-run."

    #and format the input df
    input_df = input_formatter(input_df, input_format, chrom_dict, context_size)

    combined_peg_dfs = []

    for i, val in input_df.iterrows():


        mut = mutation(val['wt_w_context'], val['alt_w_context'], val['left_context'], val['right_context'], val['Variant_Type'], val['REF'], val['ALT'], chrom=None, genome=None)
        mut.PAM_idx_forward, mut.PAM_idx_rc = eligible_PAM_finder(mut, PAM, max(RTT_lengths), proto_size)

        peg_df_plus = pegRNA_generator(mut, PAM, '+', proto_size, RTT_lengths, PBS_lengths)
        peg_df_minus = pegRNA_generator(mut, PAM, '-', proto_size, RTT_lengths, PBS_lengths)

        #need to translate PAM indexing for the - strand!!!!
        #convert to genome coordinates as well...
        peg_df = pd.concat((peg_df_plus, peg_df_minus))
        peg_df['mutation_idx'] = i

        combined_peg_dfs.append(peg_df)

    peg_df = pd.concat(combined_peg_dfs).set_index('mutation_idx').reset_index()

    if len(peg_df)==0:
        print('No pegRNAs generated; try increasing the size of the RTT in parameters')
        return peg_df
    
    #COMBINE WITH Input dataframe information...
    peg_df = pd.merge(input_df, peg_df, on='mutation_idx')

    #filter to exclude pegRNAs with no RHA (or whatever the input parameters are)
    peg_df = peg_df[peg_df['RHA_size']>=min_RHA_size]

    #do the pegRNA scoring
    peg_df = ontarget_score(peg_df)

    #calculate the peg_score
    peg_df = peggscore2(peg_df)

    #and calculate the RF score
    peg_df = RF_score(peg_df)

    #and then do the ranking
    uniq_muts = np.unique(peg_df['mutation_idx'])
    for mm in uniq_muts:
        subset = peg_df[peg_df['mutation_idx']==mm]
        subset = subset.sort_values(by=rankby, ascending=False)
        idxs = subset.index
        peg_df.loc[idxs, 'pegRNA_rank'] = list(range(1,len(subset)+1))

    if pegRNAs_per_mut not in ['All', 'all', ' All', ' all', 'all ', 'All ']:
        #filter
        peg_df = peg_df[peg_df['pegRNA_rank']<=pegRNAs_per_mut]

    #create sensor if desired
    if sensor == True:
        peg_df = sensor_generator(peg_df, proto_size, before_proto_context, sensor_length, sensor_orientation)
    
    #add in information about polyT sequences and RE sites
    peg_df = other_filtration(peg_df, RE_sites, polyT_threshold)

    #resetting the index and sorting by mutation
    peg_df = peg_df.sort_values(by=['mutation_idx', 'pegRNA_rank'], ascending=[True, True]).set_index('mutation_idx').reset_index()
    if 'index' in peg_df.keys():
        peg_df = peg_df.drop(columns='index')



    #restore columns to boolean
    peg_df['PAM_disrupted'] = peg_df['PAM_disrupted'].astype(bool)
    peg_df['Proto_disrupted'] = peg_df['Proto_disrupted'].astype(bool)

    #and add names...
    peg_df['pegRNA_id'] = ['pegRNA_' + str(i) for i in range(len(peg_df))]

    return peg_df


def peggscore2(df):
    """ 
    Function for calculating the PEGG2 Score. A multiple linear regression score based on pegRNA parameters.

    Parameters
    -----------
    df
        *type = pd.DataFrame*

        Dataframe containing the pegRNAs generated by run().

    """
    
    df['Proto_disrupted'] = np.array(df['Proto_disrupted'])+0
    df['PAM_disrupted'] = np.array(df['PAM_disrupted'])+0

    #also add in the PBS and RTT GC content
    #and the size of the variant
    pbs_gc = []
    rtt_gc = []
    var_size = []

    for i, val in df.iterrows():
        RTT = val['RTT']
        PBS = val['PBS']
        gc_count = RTT.count('G') + RTT.count('C')
        rtt_gc.append(gc_count/len(RTT))

        gc_count2 = PBS.count('G') + PBS.count('C')
        pbs_gc.append(gc_count2/len(PBS))

        ref = val['REF']
        alt = val['ALT']
        indelsize = abs(len(ref)-len(alt))
        var_size.append(indelsize)

    df['PBS_GC_content'] = pbs_gc
    df['RTT_GC_content'] = rtt_gc
    df['indel_size'] = var_size

    #check if there are any none value types in Ontarget score; if so set to 0
    df2 = df.copy()
    df2.loc[df2['OnTarget_Azimuth_Score'].isna(), 'OnTarget_Azimuth_Score'] = 0

    #dummy weights for PEGG Score
    factor_weights = {'Proto_disrupted':5, 
               'PAM_disrupted':3, 
               'indel_size':0, 
               'PBS_GC_content':0, 
               'RTT_GC_content':0, 
               'RTT_length':0, 
               'PBS_length':0, 
               'Distance_to_nick':-1, 
               'RHA_size':2, 
               'OnTarget_Azimuth_Score':.2}


    fw = [ 2.89996998,  1.58522483, -0.51088186,  8.78468344,  6.30826281,
       -0.33821302,  0.1058144 ,  0.30468214,  0.36783961,  0.14474713]

    factor_weights = dict(zip(factor_weights.keys(), fw))

    factor_lists = []
    for i in factor_weights.keys():
        weight = factor_weights[i]
        factor_lists.append(weight*np.asarray(df2[i]))

    pegg2_score = np.zeros(len(factor_lists[0]))
    for i in factor_lists:
        pegg2_score+=i
    
    df['PEGG2_Score']=pegg2_score

    return df

def RF_score(df):
    """ 
    Function for calculating the Random Forest score from pegRNA parameters.

    Parameters
    -----------
    df
        *type = pd.DataFrame*

        Dataframe containing the pegRNAs generated by run().
    """
    m = pickle.load(open('RF_PE.pkl', 'rb'))

    factors = ['Proto_disrupted', 'PAM_disrupted', 'indel_size', 'PBS_GC_content', 'RTT_GC_content', 'RTT_length', 'PBS_length', 'Distance_to_nick', 'RHA_size', 'OnTarget_Azimuth_Score']

    #check if there are any none value types in Ontarget score; if so set to 0
    df2 = df.copy()
    df2.loc[df2['OnTarget_Azimuth_Score'].isna(), 'OnTarget_Azimuth_Score'] = 0

    X = df2[factors]

    pred = m.predict(X)

    df['RF_Score'] = pred
    
    return df


#--------visualization tools-----------
def split(word):
    """
    Simple function for splitting string into component characters
    """
    return [char for char in word]

def sensor_viz(df_w_sensor, i):    
    """ 
    Function for visualizing the pegRNA aligned to the sensor. Input = dataframe with pegRNA-sensor pairs
    and the row index (i) to visualize.

    Parameters
    -----------
    df_w_sensor
        *type = pd.DataFrame*

        Dataframe containing the pegRNA-sensor pairs generated by run().

    i
        *type = int*

        Row index from the dataframe that you want to visualize.
    """
    sensor_orientation = df_w_sensor.iloc[i]['sensor_orientation']

    PAM = df_w_sensor.iloc[i]['PAM']
    PAM_len = len(PAM)

    sensor = df_w_sensor.iloc[i]['sensor_wt']
    sensor_comp = str(Bio.Seq.Seq(sensor).complement())
    split_test = split(sensor)
    split_test_comp = split(str(sensor_comp))

    if sensor_orientation == 'reverse-complement':
        #protospacer
        proto = df_w_sensor.iloc[i]['Protospacer']
        proto_start = sensor_comp.find(proto[::-1][:19])
        proto_left = ['-']*(proto_start)
        proto_right = ['-']*(len(sensor)-proto_start-len(proto))
        split_p = split(proto[::-1])
        split_proto = proto_left + split_p + proto_right

        #RTT_PBS
        RTT_len = df_w_sensor.iloc[i]['RTT_length']
        PBS_len =  df_w_sensor.iloc[i]['PBS_length']
        RTT_PBS = df_w_sensor.iloc[i]['RTT_PBS']
        RTT_PBS_end = proto_start+3+PBS_len
        RTT_PBS_start = RTT_PBS_end - len(RTT_PBS)
        left = ['-']*RTT_PBS_start
        right = ['-']*(len(sensor) - RTT_PBS_end)
        split_RTTPBS = split(RTT_PBS)
        split_RTT_PBS = left + split_RTTPBS + right

        #translation for coloring
        dict_bases = {'T':0, 'A':1, 'C':2, 'G':3, '-':4, 'X':5}
        num_translation = [dict_bases[i] for i in split_test]
        num_translation_c = [dict_bases[i] for i in split_test_comp]
        num_translation_proto = [dict_bases[i] for i in split_proto]
        num_translation_RTT = [dict_bases[i] for i in split_RTT_PBS]

        text_df = [split_proto, split_test, split_test_comp, split_RTT_PBS]
        dataFrame = [num_translation_proto, num_translation, num_translation_c,num_translation_RTT]

        fig_height = 4.5
        fig = plt.figure(figsize=(20,fig_height))
        myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white')
        #if var_type=='DEL':
        #    myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white', 'black')

        cmap = LinearSegmentedColormap.from_list('Custom', myColors, len(myColors))

        ax = sns.heatmap(dataFrame, annot=text_df, fmt="", linewidth=0, cmap=cmap, cbar=False,linewidths=0, linecolor='lightgray', xticklabels=False)

        ax.add_patch(patches.Rectangle((proto_start-PAM_len, 2), PAM_len, 1, fill=False, edgecolor='tab:green', lw=3, label='PAM')) #protospacer loca)) #protospacer location

        #protospacer
        ax.add_patch(patches.Rectangle((proto_start, 0), len(proto), 1, fill=False, edgecolor='tab:blue', lw=3, label='Protospacer')) #protospacer location

        #PBS
        ax.add_patch(patches.Rectangle((proto_start+3, 3), PBS_len, 1, fill=False, edgecolor='tab:purple', lw=3, label='PBS')) #protospacer location

        #RTT
        ax.add_patch(patches.Rectangle((proto_start+3-RTT_len, 3), RTT_len, 1, fill=False, edgecolor='yellow', lw=3, label='RTT')) #protospacer location
        ax.add_patch(Polygon([(proto_start+3-RTT_len, 3), (proto_start+3-RTT_len-2, 3.5), (proto_start+3-RTT_len, 4)],facecolor='black',edgecolor='yellow', lw=3))

        #finally, mark the location of the sequence to be mutated
        RHA = df_w_sensor.iloc[i]['RHA_size']
        vt = df_w_sensor.iloc[i]['Variant_Type']
        mut_start = proto_start+3-RTT_len + RHA

        if vt == 'INS':
            #mut_size = 2
            mut_size = len(df_w_sensor.iloc[i]['ALT'])

            ax.add_patch(patches.Rectangle((mut_start+mut_size, 1), 0, 2, fill=False, edgecolor='tab:red', lw=3, label='Mutated Bases')) #protospacer location

        elif vt=='DEL':
            mut_size = len(df_w_sensor.iloc[i]['REF'])

            ax.add_patch(patches.Rectangle((mut_start-mut_size, 1), mut_size, 2, fill=False, edgecolor='tab:red', lw=3, label='Mutated Bases')) #protospacer location


        else:
            mut_size = len(df_w_sensor.iloc[i]['REF'])

            ax.add_patch(patches.Rectangle((mut_start, 1), mut_size, 2, fill=False, edgecolor='tab:red', lw=3, label='Mutated Bases')) #protospacer location

        #also label the RHA

        ax.add_patch(patches.Rectangle((proto_start+3-RTT_len, 3), RHA, 1, fill=False, edgecolor='tab:orange', lw=3, label='RHA')) #protospacer location

    elif sensor_orientation == 'forward':

        #protospacer
        proto = df_w_sensor.iloc[i]['Protospacer']
        proto_start = sensor.find(proto[1:])-1 #account for G+19 or G+20
        proto_left = ['-']*(proto_start)
        proto_right = ['-']*(len(sensor)-proto_start-len(proto))
        split_p = split(proto)
        split_proto = proto_left + split_p + proto_right

        #RTT_PBS
        RTT_len = df_w_sensor.iloc[i]['RTT_length']
        PBS_len =  df_w_sensor.iloc[i]['PBS_length']
        RTT_PBS = str(df_w_sensor.iloc[i]['RTT_PBS'])[::-1] #[::-1] #reverse it for 3' to

        RTT_PBS_end = proto_start+3+PBS_len
        #RTT_PBS_start = RTT_PBS_end - len(RTT_PBS)

        RTT_PBS_start = proto_start + len(proto) -3 - PBS_len
        RTT_PBS_end = proto_start + len(proto) -3 + RTT_len

        left = ['-']*RTT_PBS_start
        right = ['-']*(len(sensor) - RTT_PBS_end)
        split_RTTPBS = split(RTT_PBS)
        split_RTT_PBS = left + split_RTTPBS + right

        #translation for coloring
        dict_bases = {'T':0, 'A':1, 'C':2, 'G':3, '-':4, 'X':5}
        num_translation = [dict_bases[i] for i in split_test]
        num_translation_c = [dict_bases[i] for i in split_test_comp]
        num_translation_proto = [dict_bases[i] for i in split_proto]
        num_translation_RTT = [dict_bases[i] for i in split_RTT_PBS]

        text_df = [split_RTT_PBS, split_test, split_test_comp, split_proto]
        dataFrame = [num_translation_RTT, num_translation, num_translation_c, num_translation_proto,]

        fig_height = 4.5
        fig = plt.figure(figsize=(20,fig_height))
        myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white')
        #if var_type=='DEL':
        #    myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white', 'black')

        cmap = LinearSegmentedColormap.from_list('Custom', myColors, len(myColors))

        ax = sns.heatmap(dataFrame, annot=text_df, fmt="", linewidth=0, cmap=cmap, cbar=False,linewidths=0, linecolor='lightgray', xticklabels=False)

        #PAM
        ax.add_patch(patches.Rectangle((proto_start+len(proto), 1), PAM_len, 1, fill=False, edgecolor='tab:green', lw=3, label='PAM')) #protospacer loca)) #protospacer location

        #protospacer
        ax.add_patch(patches.Rectangle((proto_start, 3), len(proto), 1, fill=False, edgecolor='tab:blue', lw=3, label='Protospacer')) #protospacer location

        #PBS
        ax.add_patch(patches.Rectangle((RTT_PBS_start, 0), PBS_len, 1, fill=False, edgecolor='tab:purple', lw=3, label='PBS')) #protospacer location

        #RTT
        ax.add_patch(patches.Rectangle((RTT_PBS_start + PBS_len, 0), RTT_len, 1, fill=False, edgecolor='yellow', lw=3, label='RTT')) #protospacer location
        ax.add_patch(Polygon([(RTT_PBS_start + PBS_len+RTT_len, 0), (RTT_PBS_start + PBS_len+RTT_len+2, 0.5), (RTT_PBS_start + PBS_len+RTT_len, 1)],facecolor='black',edgecolor='yellow', lw=3))

        #finally, mark the location of the sequence to be mutated
        RHA = df_w_sensor.iloc[i]['RHA_size']
        vt = df_w_sensor.iloc[i]['Variant_Type']
        mut_start = RTT_PBS_start + RTT_len + PBS_len - RHA -1

        if vt == 'INS':
            #mut_size = 2
            mut_size = len(df_w_sensor.iloc[i]['ALT'])

            ax.add_patch(patches.Rectangle((mut_start-mut_size+1, 1), 0, 2, fill=False, edgecolor='tab:red', lw=3, label='Mutated Bases')) #protospacer location

        elif vt=='DEL':
            mut_size = len(df_w_sensor.iloc[i]['REF'])

            ax.add_patch(patches.Rectangle((mut_start-mut_size, 1), mut_size, 2, fill=False, edgecolor='tab:red', lw=3, label='Mutated Bases')) #protospacer location


        else:
            mut_size = len(df_w_sensor.iloc[i]['REF'])

            ax.add_patch(patches.Rectangle((mut_start-mut_size+1, 1), mut_size, 2, fill=False, edgecolor='tab:red', lw=3, label='Mutated Bases')) #protospacer location

        #also label the RHA

        ax.add_patch(patches.Rectangle((mut_start+1, 0), RHA, 1, fill=False, edgecolor='tab:orange', lw=3, label='RHA')) #protospacer location

    ax.set_yticklabels(["3'", "5'", "3'", "5'"], rotation=0, fontsize=14)

    ax.legend(bbox_to_anchor=(1.16,.78), fontsize=14)

    ref_allele = df_w_sensor.iloc[i]['REF']
    mut_allele = df_w_sensor.iloc[i]['ALT']

    ax.set_title('RTT length: '+str(RTT_len)+ ' nt, PBS length: ' + str(PBS_len)+' nt | ' + vt + ': '+ref_allele + '>' + mut_allele, fontsize=16)


#----Cloning---------
def goldengate_oligos(peg_df):
    """ 
    Currently a place-holder. Going to implement automated Golden Gate oligo generator (so scaffold doesn't need to be synthesized).
    """
    return None

def prime_oligo_generator(peg_df, epeg=True, epeg_motif = 'tevopreQ1',
                    five_prime_adapter = 'AGCGTACACGTCTCACACC',three_prime_adapter = 'GAATTCTAGATCCGGTCGTCAAC',
                gRNA_scaff = 'GTTTTAGAGCTAGAAATAGCAAGTTAAAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGGCACCGAGTCGGTGC'):

    """
    A tool for automatically generating oligos from the output of run().
    Returns input dataframe with new columns containing the pegRNA oligo.

    Parameters
    ----------
    peg_df
        *type = pd.DataFrame*
        
        A dataframe containing the pegRNAs for the selected input mutations. Generated by run() or pegRNA_generator()

    epeg
        *type = bool*

        True/False whether to include an epegRNA motif at the end of the 3' extension.

    epeg_motif
        *type = str*

        Which epeg motif to include at end of 3' extension. Default = "tevopreQ1"; Other option = "mpknot" (both from Chen et al.).
        Can also input a custom motif sequence (just put in the sequence in 5' to 3' orientation)!
    
    five_prime_adapter
        *type = str*
        
        5' Prime Adapter. The automatically provided 5' adapter contains an Esp3I (BsmBI) site. Can be swapped with 
        whatever input string user wants.
    
    three_prime_adapter
        *type = str*
        
        5' Prime Adapter. The automatically provided 5' adapter contains an Esp3I (BsmBI) site. Can be swapped with 
        whatever input string user wants.
        
    gRNA_scaff
        *type = str*
        
        gRNA scaffold region. Automatically set to a functional gRNA scaffold. Can be swapped with 
        whatever input string user wants.
    
    """
        
        
    u6_term = 'TTTTTTT'
    tevopreQ1 = 'CGCGGTTCTATCTAGTTACGCGTTAAACCAACTAGAA'
    mpknot = "GGGTCAGGAGCCCCCCCCCTGAACCCAGGATAACCCTCAAAGTCGGGGGGCAACCC"

    if epeg==True:
        if epeg_motif == 'tevopreQ1':
            epeg_motif = tevopreQ1

        elif epeg_motif == 'mpknot':
            epeg_motif = mpknot

        else:
            epeg_motif=epeg_motif
            for i in epeg_motif:
                assert i in ['A','T','C','G','a','t','c','g'], "Choose a valid epeg_motif. Your custom epeg_motif contains non-DNA bases."

    elif epeg==False:
        epeg_motif = ''

    peg_oligos = []
    #check if sensor is in the peg_df
    if "sensor_wt" in peg_df.keys():
        for i, val in peg_df.iterrows():
            proto = val['Protospacer']
            extension = val['RTT_PBS']
            sensor = val["sensor_wt"]
            if sensor==None: #there are some sensor errors where it is =None
                sensor = ''

            pegRNA_full = five_prime_adapter + proto + gRNA_scaff + extension + epeg_motif + u6_term + sensor + three_prime_adapter
            peg_oligos.append(pegRNA_full)

    #if it isn't:
    else:
        for i, val in peg_df.iterrows():
            proto = val['Protospacer']
            extension = val['RTT_PBS']

            pegRNA_full = five_prime_adapter + proto + gRNA_scaff + extension + epeg_motif + u6_term + three_prime_adapter
            peg_oligos.append(pegRNA_full)


    peg_df['pegRNA_oligo'] = peg_oligos
    
    return peg_df


from pegg.prime import *
import numpy as np
import pandas as pd
import Bio.Seq
from matplotlib.colors import LinearSegmentedColormap
import seaborn as sns
import matplotlib.patches as patches
from matplotlib.patches import Polygon
import matplotlib.pyplot as plt

def eligible_PAM_finder_base(mut, PAM, proto_size=19):
    """ 
    Determines eligible PAM sequences for creating gRNAs.
    Returns PAM sequence locations on (1) forward, and (2) reverse-complement strand

    Parameters
    ------------
    mut
        *type = mutation class*

        See class: mutation

    PAM
        *type = str*
        
        PAM sequence for searching. Formatting: e.g. "NGG"

        N = [A|T|C|G]
        R = [A|G]
        Y = [C|T]
        S = [G|C]
        W = [A|T]
        K = [G|T]
        M = [A|C]
        B = [C|G|T]
        D = [A|G|T]
        H = [A|C|T]
        V = [A|C|G]

    proto_size
        *type = int*

        Size of protospacer being used. Default = 19 (G+19).
    """

    forward = pegg2.PAM_finder(mut.wt_forward, PAM)
    reverse = pegg2.PAM_finder(mut.wt_rc, PAM)

    var_type = mut.variant_type

    #first do the forward sequence
    left_len_F = len(mut.left_seq)
    left_len_R = len(mut.left_seq_rc)


    eligible_PAM_start_F = left_len_F + 1  #
    eligible_PAM_end_F = left_len_F + 1 + 19 
    
    eligible_PAM_start_R = left_len_R + 1
    eligible_PAM_end_R = left_len_R + 1 + 19 

    eligible_PAMS_F = []
    eligible_PAMS_R = []

    for i in forward:
        #check start of PAM site
        s = i[0]
        if (s>=eligible_PAM_start_F) and (s<=eligible_PAM_end_F):
            eligible_PAMS_F.append(i)

    for i in reverse:
        #check start of PAM site
        s = i[0]
        if (s>=eligible_PAM_start_R) and (s<=eligible_PAM_end_R):
            eligible_PAMS_R.append(i)

    return eligible_PAMS_F, eligible_PAMS_R


def gRNA_generator(mut, PAM, orientation, proto_size, ideal_edit_window = [4,8]):
    """ 
    Generates pegRNAs for a given mutation.

    Parameters
    ------------
    mut
        *type = mutation class*

        See class: mutation

    PAM
        *type = str*
        
        PAM sequence for searching. Formatting: e.g. "NGG"

        N = [A|T|C|G]
        R = [A|G]
        Y = [C|T]
        S = [G|C]
        W = [A|T]
        K = [G|T]
        M = [A|C]
        B = [C|G|T]
        D = [A|G|T]
        H = [A|C|T]
        V = [A|C|G]

    orientation
        *type = str*

        '+' or '-'

    proto_size
        *type = int*

        Size of protospacer being used. Default = 19 (G+19).

    ideal_edit_window
        *type = list*

        Ideal editing window for the editor being used. Default = [4,8]. Labels mutations that fall in this window for 
        future filtration if desired.
    """

    #------function----
    protospacer_seqs = []
    protospacer_wide_seqs = []
    PAM_start_list = []
    PAM_end_list = []
    PAM_list = []
    proto_location = []
    ideal_edit = []

    #Get mutation info, depending on whether we're looking on the + or the - strand
    if orientation == '+':
        left_len = len(mut.left_seq)
        ref_len = len(mut.ref_seq)
        alt_len = len(mut.alt_seq)
        alt_seq = mut.alt_seq
        seq_F = mut.wt_forward

        PAM_seqs = mut.PAM_idx_forward

    elif orientation == '-':
        left_len = len(mut.left_seq_rc)
        ref_len = len(mut.ref_seq_rc)
        alt_len = len(mut.alt_seq_rc)
        alt_seq = mut.alt_seq_rc
        seq_F = mut.wt_rc

        PAM_seqs = mut.PAM_idx_rc

    #iterate over the PAM sequences:
    for i in PAM_seqs:
        PAM_start = i[0]
        PAM_end = i[1]

        proto_loc = PAM_start - left_len
        #and then convert to 1 to 20 indexing...
        proto_loc = 21-proto_loc
        proto_location.append(proto_loc)
        if (proto_loc>=ideal_edit_window[0]) and (proto_loc<=ideal_edit_window[0]):
            ideal_edit.append(True)
        else:
            ideal_edit.append(False)

    #pull out the PAM sequence and protospacer as well
        PAM_sequence = seq_F[PAM_start:PAM_end]
        protospacer =  'G' + seq_F[PAM_start - proto_size:PAM_start]

        #extract wide protospacer for Rule set 2 prediction...
        protospacer_wide = seq_F[PAM_start - 20 - 4 : PAM_start + 3 + 3]

        #------add in PAM info, protospacer, + PBS/RTT_PBS
        #record info about PAM sequences
        if orientation == '+':
            PAM_start_list.append(PAM_start)
            PAM_end_list.append(PAM_end)
        elif orientation == '-':
            PAM_start_list.append(len(seq_F)-PAM_start)
            PAM_end_list.append(len(seq_F)-PAM_end)

        PAM_list.append(PAM_sequence)
        protospacer_seqs.append(protospacer)
        protospacer_wide_seqs.append(protospacer_wide)

    cols = [PAM_start_list, PAM_end_list, PAM_list, orientation, protospacer_wide_seqs, protospacer_seqs, proto_location, ideal_edit]
    col_labels = ["PAM_start", "PAM_end", "PAM", "PAM_strand","Protospacer_30", "Protospacer", "Protospacer_Location", "Ideal_Edit_Window"]
    dtypes = ['int', 'int', 'str', 'str', 'str', 'str', 'int', 'bool']

    dtype_dict = dict(zip(col_labels, dtypes))

    peg_df = pd.DataFrame(dict(zip(col_labels, cols))).dropna().reset_index().drop(columns='index').astype(dtype_dict)

    return peg_df



def run_base(input_df, input_format, chrom_dict, PAM = "NGG", filtration = "ABE+CBE", ideal_edit_window = [4,8], auto_SNP_filter = True, 
        proto_size=19, context_size = 120, 
        RE_sites=None, polyT_threshold=4, 
        before_proto_context=5, sensor_length=40, sensor_orientation = 'reverse-complement', sensor=True):
            
    """ 
    Master function for generating base editing gRNAs. Takes as input a dataframe containing mutations in one of the acceptable formats.
    Returns a dataframe with gRNAs with desired design parameters.

    Parameters
    ------------
    input_df
        *type = pd.DataFrame*

        Pandas dataframe that contains the input mutations.

    input_format
        *type = str*

        Options = 'cBioPortal', 'WT_ALT', 'PrimeDesign'. For 'WT_ALT', make sure to put headers as "WT" and "ALT". 
        For "PrimeDesign", put the header as "SEQ".

    chrom_dict
        *type = dict or None*

        Dictionary generated by genome_loader() that holds the chromosome sequences.

    PAM
        *type = str*

        PAM sequence for searching. Default = "NGG". Can include any nucleic acid code (e.g. PAM = "NRCH").

    filtration
        *type = str or list*

        Filters the mutation input list to only include the desired SNPs. Options = "No filter", "ABE", "CBE", "ABE+CBE",
        or a list containing the desired SNPs to model (e.g. ['C>A', 'T>C']).

    ideal_edit_window
        *type = list*

        Ideal editing window for the editor being used. Default = [4,8]. Labels mutations that fall in this window for 
        future filtration if desired.

    auto_SNP_filter
        *type = bool*

        True/False for whether to filter mutant input to exclude mutations that are NOT SNPs (and thus not BE amenable).
    
    proto_size
        *type = int*

        The length of the protospacer (excluding the appended G at the 5' end). Default = 19 (G+19).

    context_size
        *type = int*

        The amount of context/flanking sequence on either side of the mutation to generate. For larger variants, set this larger.
        Default = 120. e.g. AAA(A/G)AAA = context_size of 3

    RE_sites
        *type = list or None*

        A list containing the RE recognition sites to filter (e.g. ['CGTCTC', 'GAATTC'] for Esp3I and EcoRI). Default = None (no filtration).

    polyT_threshold
        *type = int*

        The length of the polyT sequence to classify as a terminator. Default = 4.

    before_proto_context
        *type = int*

        Default = 5. Amount of nucleotide context to put before the protospacer in the sensor

    sensor_length
        *type = int*

        Total length of the sensor in nt. Default = 60. 

    sensor_orientation
        *type = str*

        Options for sensor_orientation = 'reverse-complement' or'forward'.

    sensor
        *type = bool*

        True/False whether to include a sensor in the pegRNA design or not. 
    """

    input_df = pegg2.input_formatter(input_df, input_format, chrom_dict, context_size)

    vts = np.unique(input_df['Variant_Type'])

    #make sure it's a SNP
    if auto_SNP_filter==False:
        not_allowed = ['ONP', 'DEL', 'INS', 'INDEL']
        for i in not_allowed:
            assert i not in vts, "Only SNPs can be modeled with base editing; remove non-SNPs from mutation list or select auto_SNP_filter=True"

    elif auto_SNP_filter==True:
        input_df = input_df[input_df['Variant_Type']=='SNP']
        not_allowed = ['ONP', 'DEL', 'INS', 'INDEL']
        for i in not_allowed:
            if i in vts:
                print("Only SNPs can be modeled with base editing; Non-SNPs automatically removed from list")

    combined_peg_dfs = []

    for i, val in input_df.iterrows():


        mut = pegg2.mutation(val['wt_w_context'], val['alt_w_context'], val['left_context'], val['right_context'], val['Variant_Type'], val['REF'], val['ALT'], chrom=None, genome=None)
        mut.PAM_idx_forward, mut.PAM_idx_rc = eligible_PAM_finder_base(mut, PAM, proto_size)

        peg_df_plus = gRNA_generator(mut, PAM, '+', proto_size, ideal_edit_window)
        peg_df_minus = gRNA_generator(mut, PAM, '-', proto_size, ideal_edit_window)

        #need to translate PAM indexing for the - strand!!!!
        #convert to genome coordinates as well...
        peg_df = pd.concat((peg_df_plus, peg_df_minus))
        peg_df['mutation_idx'] = i

        combined_peg_dfs.append(peg_df)

    peg_df = pd.concat(combined_peg_dfs).set_index('mutation_idx').reset_index()

    #COMBINE WITH Input dataframe information...
    peg_df = pd.merge(input_df, peg_df, on='mutation_idx')

    #and then do filtration...
    if filtration != 'No filter':
        if filtration == 'CBE':
            p1 = peg_df[(peg_df['REF']=='C') & (peg_df['ALT']=='T') & (peg_df['PAM_strand']=='+')]
            p2 = peg_df[(peg_df['REF']=='G') & (peg_df['ALT']=='A') & (peg_df['PAM_strand']=='-')]
            peg_df = pd.concat((p1, p2))
            peg_df['Editor'] = 'CBE'

        elif filtration == 'ABE':
            p1 = peg_df[(peg_df['REF']=='A') & (peg_df['ALT']=='G') & (peg_df['PAM_strand']=='+')]
            p2 = peg_df[(peg_df['REF']=='T') & (peg_df['ALT']=='C') & (peg_df['PAM_strand']=='-')]
            peg_df = pd.concat((p1, p2))
            peg_df['Editor'] = 'ABE'

        elif filtration in ['ABE+CBE', 'CBE+ABE', 'ABE + CBE', 'CBE + ABE']:
            #filter for only ABE/CBE eligible variants and PAM strand...
            p1 = peg_df[(peg_df['REF']=='C') & (peg_df['ALT']=='T') & (peg_df['PAM_strand']=='+')]
            p2 = peg_df[(peg_df['REF']=='G') & (peg_df['ALT']=='A') & (peg_df['PAM_strand']=='-')]
            peg_df1 = pd.concat((p1, p2))
            peg_df1['Editor'] = 'CBE'

            p1 = peg_df[(peg_df['REF']=='A') & (peg_df['ALT']=='G') & (peg_df['PAM_strand']=='+')]
            p2 = peg_df[(peg_df['REF']=='T') & (peg_df['ALT']=='C') & (peg_df['PAM_strand']=='-')]
            peg_df2 = pd.concat((p1, p2))
            peg_df2['Editor'] = 'ABE'

            peg_df = pd.concat((peg_df1, peg_df2))

        elif type(filtration)==list:

            p_holder = []
            for k in filtration:
                spl = k.split('>')
                r = spl[0]
                r_rc = str(Bio.Seq.Seq(r).complement())
                a = spl[1] 
                a_rc = str(Bio.Seq.Seq(a).complement())

                p1 = peg_df[(peg_df['REF']==r) & (peg_df['ALT']==a) & (peg_df['PAM_strand']=='+')]
                p2 = peg_df[(peg_df['REF']==r_rc) & (peg_df['ALT']==a_rc) & (peg_df['PAM_strand']=='-')]
                peg_df2 = pd.concat((p1, p2))
                peg_df2['Editor'] = k
                p_holder.append(peg_df2)

            peg_df = pd.concat(p_holder)

        else:
            assert 1==2, "Choose valid filtration format ('ABE', 'CBE', 'ABE+CBE', or list format (['C>G', ...]))"


    #resetting the index
    peg_df = peg_df.sort_values(by='mutation_idx', ascending=True).set_index('mutation_idx').reset_index()
    if 'index' in peg_df.keys():
        peg_df = peg_df.drop(columns='index')

    #and add names...
    peg_df['gRNA_id'] = ['gRNA_' + str(i) for i in range(len(peg_df))]

    #do the pegRNA scoring
    #first add in the on-target score...
    #MAYBE LIMIT THIS TO ONLY NGG PAMS???
    peg_df = pegg2.ontarget_score(peg_df)

    #then calculate the PEGG Score and filter out the pegRNAs...
    #have an option for "All" possible pegRNAs with given parameters

    #create sensor if desired
    if sensor == True:
        peg_df = sensor_generator_base(peg_df, proto_size, before_proto_context, sensor_length, sensor_orientation)

    #add in information about polyT sequences and RE sites
    peg_df = pegg2.other_filtration(peg_df, RE_sites, polyT_threshold)
    
    return peg_df


def sensor_generator_base(df, proto_size, before_proto_context=5, sensor_length=60, sensor_orientation = 'reverse-complement'):
    """ 
    Generates sensor sequence for quantification of gRNA editing outcomes.
    Automatically puts sensor in reverse complement orientation with respect to protospacer.
    This is highly reccomended to reduce recombination during cloning/library preparation.

    Parameters
    -----------
    df
        *type = pd.DataFrame*

        Dataframe containing gRNAs.

    proto_size
        *type = int*

        Size of protospacer. 

    before_proto_context
        *type = int*

        Default = 5. Amount of nucleotide context to put before the protospacer in the sensor

    sensor_length
        *type = int*

        Total length of the sensor in nt. Default = 60. 

    sensor_orientation
        *type = str*

        Options for sensor_orientation = 'reverse-complement' or'forward'.

    """

    assert sensor_orientation in ['reverse-complement', 'forward'], "Choose an eligible sensor orientation option ('reverse_comp' (default) or 'forward')"

    sensor_error = []
    sensor_wt_seqs = []
    sensor_alt_seqs = []

    for i, val in df.iterrows():
        wt = val['wt_w_context']
        alt = val['alt_w_context']
        ref_minus_alt = len(val['REF']) - len(val['ALT'])
        strand = val['PAM_strand']
        pam_start = val['PAM_start']
        pam_end = val['PAM_end']
        #RTT_PBS = val['RTT_PBS']
        #RHA = val['RHA_size']

        #if PAM is on negative strand; flip it
        if strand == '-':
            pam_start = len(wt) - pam_start
            pam_end = len(wt) - pam_end
            wt = str(Bio.Seq.Seq(wt).reverse_complement())
            alt = str(Bio.Seq.Seq(alt).reverse_complement())

        s1 = pam_start-(proto_size+1)-before_proto_context
        if s1 <0:
            sensor_wt_seqs.append(None)
            sensor_alt_seqs.append(None)
            sensor_error.append('insufficient context sequence provided; try increasing context size')

        else:
            proto_pam = wt[s1:pam_end]
            proto_pam_alt = alt[s1:pam_end]

            remain_len = sensor_length - len(proto_pam)
            if remain_len < 0:
                sensor_wt_seqs.append(None)
                sensor_alt_seqs.append(None)
                sensor_error.append('sensor too small; increase sensor size in parameters; or decrease before_proto_context')

            else:
                remain_len_alt = sensor_length - len(proto_pam) - ref_minus_alt

                e1 = pam_end + remain_len
                e2 = pam_end + remain_len_alt
                if e1>len(wt):
                    sensor_wt_seqs.append(None)
                    sensor_alt_seqs.append(None)
                    sensor_error.append('insufficient context sequence provided; try increasing context size')

                else:
                    rtt_region = wt[pam_end: e1]
                    rtt_region_alt = alt[pam_end:e2]

                    sensor_wt = proto_pam + rtt_region
                    sensor_alt = proto_pam_alt + rtt_region_alt

                    #AND FINALLY either keep it as is, or take the reverse complement
                    if sensor_orientation == 'reverse-complement':
                        sensor_wt = str(Bio.Seq.Seq(sensor_wt).reverse_complement())
                        sensor_alt = str(Bio.Seq.Seq(sensor_alt).reverse_complement())

                   
                        sensor_wt_seqs.append(sensor_wt)
                        sensor_alt_seqs.append(sensor_alt)
                        sensor_error.append("No Error")
                    
                    if sensor_orientation == 'forward':
                    
                        sensor_wt_seqs.append(sensor_wt)
                        sensor_alt_seqs.append(sensor_alt)
                        sensor_error.append("No Error")


    df['sensor_wt'] = sensor_wt_seqs
    df['sensor_alt'] = sensor_alt_seqs
    df['sensor_orientation'] = sensor_orientation
    df['sensor_error'] = sensor_error

    return df

#--------Visualization tools----------
def split(word):
    """
    Simple function for splitting string into component characters
    """
    return [char for char in word]

def sensor_viz_base(df_w_sensor, i):    
    """ 
    Function for visualizing the gRNA aligned to the sensor. Input = dataframe with pegRNA-sensor pairs
    and the row index (i) to visualize.

    Parameters
    -----------
    df_w_sensor
        *type = pd.DataFrame*

        Dataframe containing the gRNA-sensor pairs generated by run().

    i
        *type = int*

        Row index from the dataframe that you want to visualize.
    """
    sensor_orientation = df_w_sensor.iloc[i]['sensor_orientation']

    PAM = df_w_sensor.iloc[i]['PAM']
    PAM_len = len(PAM)

    sensor = df_w_sensor.iloc[i]['sensor_wt']
    sensor_comp = str(Bio.Seq.Seq(sensor).complement())
    split_test = split(sensor)
    split_test_comp = split(str(sensor_comp))

    proto_loc = df_w_sensor.iloc[i]['Protospacer_Location']

    if sensor_orientation == 'reverse-complement':
        #protospacer
        proto = df_w_sensor.iloc[i]['Protospacer']
        proto_start = sensor_comp.find(proto[::-1][:19])
        proto_left = ['-']*(proto_start)
        proto_right = ['-']*(len(sensor)-proto_start-len(proto))
        split_p = split(proto[::-1])
        split_proto = proto_left + split_p + proto_right

        #translation for coloring
        dict_bases = {'T':0, 'A':1, 'C':2, 'G':3, '-':4, 'X':5}
        num_translation = [dict_bases[i] for i in split_test]
        num_translation_c = [dict_bases[i] for i in split_test_comp]
        num_translation_proto = [dict_bases[i] for i in split_proto]

        text_df = [split_proto, split_test, split_test_comp]
        dataFrame = [num_translation_proto, num_translation, num_translation_c]

        fig_height = 4.5
        fig = plt.figure(figsize=(20,fig_height))
        myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white')
        #if var_type=='DEL':
        #    myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white', 'black')

        cmap = LinearSegmentedColormap.from_list('Custom', myColors, len(myColors))

        ax = sns.heatmap(dataFrame, annot=text_df, fmt="", linewidth=0, cmap=cmap, cbar=False,linewidths=0, linecolor='lightgray', xticklabels=False, yticklabels=True)
        ax.set_yticklabels(["3'", "5'", "3'"], rotation=0, fontsize=14)


        ax.add_patch(patches.Rectangle((proto_start-PAM_len, 2), PAM_len, 1, fill=False, edgecolor='tab:green', lw=3, label='PAM')) #protospacer loca)) #protospacer location

        #protospacer
        ax.add_patch(patches.Rectangle((proto_start, 0), len(proto), 1, fill=False, edgecolor='tab:blue', lw=3, label='Protospacer')) #protospacer location

        #mutant location
        mut_start = proto_start+(20-proto_loc)
        print(mut_start)
        ax.add_patch(patches.Rectangle((mut_start, 2), 1, 1, fill=False, edgecolor='tab:red', lw=3, label='Mutated Base')) #protospacer location

    elif sensor_orientation == 'forward':

        #protospacer
        proto = df_w_sensor.iloc[i]['Protospacer']
        proto_start = sensor.find(proto[1:])-1 #account for G+19 or G+20
        proto_left = ['-']*(proto_start)
        proto_right = ['-']*(len(sensor)-proto_start-len(proto))
        split_p = split(proto)
        split_proto = proto_left + split_p + proto_right

        
        #translation for coloring
        dict_bases = {'T':0, 'A':1, 'C':2, 'G':3, '-':4, 'X':5}
        num_translation = [dict_bases[i] for i in split_test]
        num_translation_c = [dict_bases[i] for i in split_test_comp]
        num_translation_proto = [dict_bases[i] for i in split_proto]

        text_df = [split_test, split_test_comp, split_proto]
        dataFrame = [num_translation, num_translation_c, num_translation_proto]

        fig_height = 4.5
        fig = plt.figure(figsize=(20,fig_height))
        myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white')
        #if var_type=='DEL':
        #    myColors = ('plum', 'skyblue', 'navajowhite', 'lightgoldenrodyellow', 'white', 'black')

        cmap = LinearSegmentedColormap.from_list('Custom', myColors, len(myColors))

        ax = sns.heatmap(dataFrame, annot=text_df, fmt="", linewidth=0, cmap=cmap, cbar=False,linewidths=0, linecolor='lightgray', xticklabels=False)

        ax.set_yticklabels(["5'", "3'", "5'"], rotation=0, fontsize=14)

        #PAM
        ax.add_patch(patches.Rectangle((proto_start+len(proto), 0), PAM_len, 1, fill=False, edgecolor='tab:green', lw=3, label='PAM')) #protospacer loca)) #protospacer location

        #protospacer
        ax.add_patch(patches.Rectangle((proto_start, 2), len(proto), 1, fill=False, edgecolor='tab:blue', lw=3, label='Protospacer')) #protospacer location

        #mutation_location
        mut_start = proto_loc+proto_start-1
        ax.add_patch(patches.Rectangle((mut_start, 0), 1, 1, fill=False, edgecolor='tab:red', lw=3, label='Mutated Base')) #protospacer location


    ax.legend(bbox_to_anchor=(1.16,.78), fontsize=14)

    ref_allele = df_w_sensor.iloc[i]['REF']
    mut_allele = df_w_sensor.iloc[i]['ALT']
    strand = df_w_sensor.iloc[i]['PAM_strand']

    ax.set_title(f'PAM strand = {strand} | ' + ref_allele + '>' + mut_allele + f" | Protospacer Location = +{proto_loc}", fontsize=16)


#------oligo generation tool
def base_oligo_generator(peg_df, five_prime_adapter = 'AGCGTACACGTCTCACACC',three_prime_adapter = 'GAATTCTAGATCCGGTCGTCAAC',
                gRNA_scaff = 'GTTTTAGAGCTAGAAATAGCAAGTTAAAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGGCACCGAGTCGGTGC'):

    """
    A tool for automatically generating oligos from the output of run().
    Returns input dataframe with new columns containing gRNA oligo with or without sensor.

    Parameters
    ----------
    peg_df
        *type = pd.DataFrame*
        
        A dataframe containing the gRNAs for the selected input mutations. Generated by run_base() or gRNA_generator()

    five_prime_adapter
        *type = str*
        
        5' Prime Adapter. The automatically provided 5' adapter contains an Esp3I (BsmBI) site. Can be swapped with 
        whatever input string user wants.
    
    three_prime_adapter
        *type = str*
        
        5' Prime Adapter. The automatically provided 5' adapter contains an Esp3I (BsmBI) site. Can be swapped with 
        whatever input string user wants.
        
    gRNA_scaff
        *type = str*
        
        gRNA scaffold region. Automatically set to a functional gRNA scaffold. Can be swapped with 
        whatever input string user wants.
    
    """
        
        
    u6_term = 'TTTTTTT'

    base_oligos = []
    #check if sensor is in the peg_df
    if "sensor_wt" in peg_df.keys():
        for i, val in peg_df.iterrows():
            proto = val['Protospacer']
            extension = val['RTT_PBS']
            sensor = val["sensor_wt"]
            if sensor==None: #there are some sensor errors where it is =None
                sensor = ''

            gRNA_full = five_prime_adapter + proto + gRNA_scaff + u6_term + sensor + three_prime_adapter
            base_oligos.append(pegRNA_full)

    #if it isn't:
    else:
        for i, val in peg_df.iterrows():
            proto = val['Protospacer']
            extension = val['RTT_PBS']

            gRNA_full = five_prime_adapter + proto + gRNA_scaff + u6_term + three_prime_adapter
            base_oligos.append(pegRNA_full)


    peg_df['gRNA_oligo'] = base_oligos
    
    return peg_df